import unittest
import xml.etree.ElementTree as ET
import pysvd


class TestNodeElement(unittest.TestCase):
    xml = '''
    <node><value>
        text
    </value></node>'''

    def test_value(self):
        node = ET.fromstring(self.xml)
        test = pysvd.node.Element(node, 'value')

        self.assertEqual(test, 'text')

    def test_none(self):
        node = ET.fromstring(self.xml)
        test = pysvd.node.Element(node, 'unknown')

        self.assertIsNone(test)

    def test_exception(self):
        node = ET.fromstring(self.xml)
        with self.assertRaises(SyntaxError):
            pysvd.node.Element(node, 'unknown', True)


class TestNodeAttribute(unittest.TestCase):
    xml = '''
    <node value=" text " />
    '''

    def test_value(self):
        node = ET.fromstring(self.xml)
        test = pysvd.node.Attribute(node, 'value')

        self.assertEqual(test, 'text')

    def test_none(self):
        node = ET.fromstring(self.xml)
        test = pysvd.node.Attribute(node, 'unknown')

        self.assertIsNone(test)

    def test_exception(self):
        node = ET.fromstring(self.xml)
        with self.assertRaises(SyntaxError):
            pysvd.node.Attribute(node, 'unknown', True)



import re


def Element(node, tag, mandatory=False):
    """Get the element text for the provided tag from the provided node"""

    value = node.findtext(tag)
    if value is None:
        if mandatory:
            raise SyntaxError("Element '{}.{}' is mandatory, but not present!".format(node.tag, tag))
        return None
    else:
        value = re.sub(r'\s+', ' ', value)
        return value.strip()


def Attribute(node, tag, mandatory=False):
    """Get the attribute text for the provided tag from the provided node"""

    value = node.get(tag)
    if value is None:
        if mandatory:
            raise SyntaxError("Attribute '{}@{}' is mandatory, but not present!".format(node.tag, tag))
        return None
    else:
        value = re.sub(r'\s+', ' ', value)
        return value.strip()



def Text(value, default=None):
    """Get text from the provided value. If None, default is returned"""

    if value is None:
        return default
    else:
        return value


def Integer(value, default=None):
    """Get integer value from the provided value.
    If None, default is returned
    """

    if value is None:
        return default

    value = value.lower()

    # Hexadecimal '0x'
    if value.startswith('0x'):
        return int(value[2:], 16)
    # Binary '0b'
    elif value.startswith('0b'):
        value = value.replace('x', '0')[2:]
        return int(value, 2)
    # Binary '#'
    elif value.startswith('#'):
        value = value.replace('x', '0')[1:]
        return int(value, 2)
    # Decimal
    else:
        return int(value)


def Boolean(value, default=None):
    """Get boolean value from the provided value.
    If None, default is returned
    """

    if value is None:
        return default

    value = value.lower()
    if value == 'false' or value == '0':
        return False
    elif value == 'true' or value == '1':
        return True
    else:
        raise ValueError("Can not convert str '{}' to boolean".format(value))


def Enum(enum, value, default=None):
    """Get enum value from the provided value looking up in EnumClass.
    If None, default is returned
    """
    if value is None:
        return default

    for pair in enum:
        if pair.value == value:
            return pair

    raise KeyError("Value '{}' not contained in enum type".format(value))



import unittest
import pysvd


class TestParserText(unittest.TestCase):

    def test_no_param(self):
        test = pysvd.parser.Text('Text')

        self.assertEqual(test, 'Text')

    def test_default_none(self):
        test = pysvd.parser.Text(None)

        self.assertIsNone(test)

    def test_default_value(self):
        test = pysvd.parser.Text(None, 'value')

        self.assertEqual(test, 'value')


class TestParserInteger(unittest.TestCase):

    def test_default_none(self):
        test = pysvd.parser.Integer(None)

        self.assertIsNone(test)

    def test_default_value(self):
        test = pysvd.parser.Integer(None, 16)

        self.assertEqual(test, 16)

    def test_decimal(self):
        test = pysvd.parser.Integer('4711')

        self.assertEqual(test, 4711)

    def test_hexadecimal(self):
        test = pysvd.parser.Integer('0x1234')

        self.assertEqual(test, 0x1234)

    def test_binary_0b(self):
        test = pysvd.parser.Integer('0b110xx011')

        self.assertEqual(test, 195)

    def test_binary_hash(self):
        test = pysvd.parser.Integer('#110xx011')

        self.assertEqual(test, 195)

    def test_exception(self):
        with self.assertRaises(ValueError):
            pysvd.parser.Integer('text')


class TestParserBoolean(unittest.TestCase):

    def test_default_none(self):
        test = pysvd.parser.Boolean(None)

        self.assertIsNone(test)

    def test_default_value(self):
        test = pysvd.parser.Boolean(None, False)

        self.assertEqual(test, False)

    def test_false_value(self):
        test = pysvd.parser.Boolean('0')

        self.assertEqual(test, False)

    def test_false_text(self):
        test = pysvd.parser.Boolean('FaLsE')

        self.assertEqual(test, False)

    def test_true_value(self):
        test = pysvd.parser.Boolean('1')

        self.assertEqual(test, True)

    def test_true_text(self):
        test = pysvd.parser.Boolean('tRuE')

        self.assertEqual(test, True)

    def test_exception(self):
        with self.assertRaises(ValueError):
            pysvd.parser.Boolean('text')


class TestParserEnum(unittest.TestCase):

    def test_default_none(self):
        test = pysvd.parser.Enum(pysvd.type.access, None)

        self.assertIsNone(test)

    def test_default_value(self):
        test = pysvd.parser.Enum(
            pysvd.type.access,
            None,
            pysvd.type.access.read_only)

        self.assertEqual(test, pysvd.type.access.read_only)

    def test_enum(self):
        test = pysvd.parser.Enum(pysvd.type.access, 'read-write')

        self.assertEqual(test, pysvd.type.access.read_write)

    def test_exception(self):
        with self.assertRaises(KeyError):
            pysvd.parser.Enum(pysvd.type.access, 'unknown')



import unittest
import xml.etree.ElementTree as ET

import pysvd


# Used in TestClassGroup.test_group_attributes
class HelperClassGroupAttributes(pysvd.classes.Group):

    attributes = ['extra']

    def __init__(self, parent, node):
        super().__init__(parent, node)


class HelperClassDeriveRoot(pysvd.classes.Base):

    def __init__(self, node):
        self.name = "root"
        self.register = []

        super().__init__(node)

    def parse(self, node):
        super().parse(node)

        HelperClassDeriveRegister.add_elements(self, self.register, node, 'register')

    def find(self, name):
        for register in self.register:
            if register.name == name:
                return register
        return None


class HelperClassDeriveRegister(pysvd.classes.Derive):

    def __init__(self, parent, node):
        self.field = []

        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'name', pysvd.parser.Text, True)
        self.add_attribute(node, 'description', pysvd.parser.Text, True)
        self.add_attribute(node, 'addressOffset', pysvd.parser.Integer)
        self.add_attribute(node, 'size', pysvd.parser.Integer)

        HelperClassDeriveField.add_elements(self, self.field, node, 'field')

    def find(self, name):
        for field in self.field:
            if field.name == name:
                return field
        return None


class HelperClassDeriveField(pysvd.classes.Derive):

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'name', pysvd.parser.Text, True)
        self.add_attribute(node, 'description', pysvd.parser.Text, True)
        self.add_attribute(node, 'access', pysvd.type.access)


class HelperClassDim(pysvd.classes.Dim):

    def __init__(self, parent, node):
        self.offset = 0
        super().__init__(parent, node)

    def set_offset(self, value):
        self.offset += value

    def parse(self, node):
        super().parse(node)


class TestClassBase(unittest.TestCase):

    def test_inheritance(self):
        self.assertTrue(issubclass(pysvd.classes.Base, object))

    def test_ctor(self):
        test = pysvd.classes.Base(None)

        self.assertEqual(type(test), pysvd.classes.Base)
        self.assertIsNone(test.node)
        self.assertIsNone(test.parent)
        self.assertIsNone(test.derivedFrom)

    def test_attributes(self):
        test = pysvd.classes.Base(None)

        attr = {}
        attr['name'] = 'test'
        attr['enable'] = True
        attr['none'] = None
        test.__dict__.update(attr)

        self.assertEqual(test.name, 'test')
        self.assertTrue(test.enable)
        with self.assertRaises(AttributeError):
            self.assertNone(test.none)

    def test_find(self):
        test = pysvd.classes.Base(None)

        self.assertIsNone(test.find("test"))


class TestClassParent(unittest.TestCase):

    def test_inheritance(self):
        self.assertTrue(issubclass(pysvd.classes.Parent, pysvd.classes.Base))

    def test_ctor(self):
        test = pysvd.classes.Parent(None, None)
        child = pysvd.classes.Parent(test, None)

        self.assertEqual(type(test), pysvd.classes.Parent)
        self.assertIsNone(test.parent)
        self.assertEqual(child.parent, test)

    def test_attributes(self):
        test = pysvd.classes.Parent(None, None)

        attr = {}
        attr['name'] = 'test'
        attr['enable'] = True
        attr['none'] = None
        test.__dict__.update(attr)

        self.assertEqual(test.name, 'test')
        self.assertTrue(test.enable)
        with self.assertRaises(AttributeError):
            self.assertNone(test.none)

    def test_find(self):
        test = pysvd.classes.Parent(None,  None)

        self.assertIsNone(test.find("test"))


class TestClassGroup(unittest.TestCase):

    def test_inheritance(self):
        self.assertTrue(issubclass(pysvd.classes.Group, pysvd.classes.Parent))

    def test_ctor(self):
        test = pysvd.classes.Group(None, None)
        child = pysvd.classes.Group(test, None)

        self.assertEqual(type(test), pysvd.classes.Group)
        self.assertIsNone(test.parent)
        self.assertEqual(child.parent, test)

    def test_attributes(self):
        test = pysvd.classes.Group(None, None)

        attr = {}
        attr['name'] = 'test'
        attr['enable'] = True
        attr['none'] = None
        test.__dict__.update(attr)

        self.assertEqual(test.name, 'test')
        self.assertTrue(test.enable)
        with self.assertRaises(AttributeError):
            self.assertNone(test.none)

    def test_group(self):
        test_attr = {
            'name': 'test',
            'size': 8,
            'resetValue': 0xFFFF,
            'extra': 'xxx',
        }

        child_attr = {
            'name': 'child',
            'size': 16,
        }

        test = pysvd.classes.Group(None, None)
        test.__dict__.update(test_attr)
        child = pysvd.classes.Group(test, None)
        child.__dict__.update(child_attr)

        self.assertEqual(type(test), pysvd.classes.Group)
        self.assertEqual(type(child), pysvd.classes.Group)

        self.assertIsNone(test.parent)
        self.assertEqual(child.parent, test)

        self.assertEqual(test.name, 'test')
        self.assertEqual(test.size, 8)
        self.assertEqual(test.resetValue, 0xFFFF)
        self.assertEqual(test.extra, 'xxx')

        self.assertEqual(child.name, 'child')
        self.assertEqual(child.size, 16)
        self.assertEqual(child.resetValue, 0xFFFF)
        with self.assertRaises(AttributeError):
            self.assertIsNone(child.extra)

    def test_group_multilevel(self):
        test_attr = {
            'name': 'test',
            'size': 8,
            'resetValue': 0xFFFF,
            'extra': 'xxx',
        }

        child_attr = {
            'name': 'child',
            'size': 16,
        }

        subchild_attr = {
            'name': 'subchild',
        }

        test = pysvd.classes.Group(None, None)
        test.__dict__.update(test_attr)
        child = pysvd.classes.Group(test, None)
        child.__dict__.update(child_attr)
        subchild = pysvd.classes.Group(child, None)
        subchild.__dict__.update(subchild_attr)

        self.assertEqual(type(test), pysvd.classes.Group)
        self.assertEqual(type(child), pysvd.classes.Group)
        self.assertEqual(type(subchild), pysvd.classes.Group)

        self.assertIsNone(test.parent)
        self.assertEqual(child.parent, test)
        self.assertEqual(subchild.parent, child)

        self.assertEqual(test.name, 'test')
        self.assertEqual(test.size, 8)
        self.assertEqual(test.resetValue, 0xFFFF)
        self.assertEqual(test.extra, 'xxx')

        self.assertEqual(child.name, 'child')
        self.assertEqual(child.size, 16)
        self.assertEqual(child.resetValue, 0xFFFF)
        with self.assertRaises(AttributeError):
            self.assertIsNone(child.extra)

        self.assertEqual(subchild.name, 'subchild')
        self.assertEqual(subchild.size, 16)
        self.assertEqual(subchild.resetValue, 0xFFFF)
        with self.assertRaises(AttributeError):
            self.assertIsNone(subchild.extra)

    def test_group_attributes(self):
        test_attr = {
            'name': 'test',
            'size': 8,
            'resetValue': 0xFFFF,
            'extra': 'xxx',
        }

        child_attr = {
            'name': 'child',
            'size': 16,
        }

        test = pysvd.classes.Group(None, None)
        test.__dict__.update(test_attr)
        child = HelperClassGroupAttributes(test, None)
        child.__dict__.update(child_attr)

        self.assertEqual(type(test), pysvd.classes.Group)
        self.assertEqual(type(child), HelperClassGroupAttributes)

        self.assertIsNone(test.parent)
        self.assertEqual(child.parent, test)

        self.assertEqual(test.name, 'test')
        self.assertEqual(test.size, 8)
        self.assertEqual(test.resetValue, 0xFFFF)
        self.assertEqual(test.extra, 'xxx')

        self.assertEqual(child.name, 'child')
        self.assertEqual(child.size, 16)
        self.assertEqual(child.extra, 'xxx')

        with self.assertRaises(AttributeError):
            self.assertIsNone(child.reset_value)


class TestClassDerive(unittest.TestCase):

    def test_inheritance(self):
        self.assertTrue(issubclass(pysvd.classes.Derive, pysvd.classes.Group))

    def test_unknown_element_exception(self):
        xml = '''
        <root>
            <register>
                <name>TimerCtrl0</name>
                <description>Timer Control Register</description>
            </register>
            <register derivedFrom="UartCtrl0">
                <name>TimerCtrl1</name>
                <description>Derived Timer</description>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        with self.assertRaises(KeyError):
            HelperClassDeriveRoot(node)

    def test_path_level_exception(self):
        xml = '''
        <root>
            <register>
                <name>TimerCtrl0</name>
                <description>Timer Control Register</description>
            </register>
            <register derivedFrom="TimerCtrl0.BitField0">
                <name>TimerCtrl1</name>
                <description>Derived Timer</description>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        with self.assertRaises(KeyError):
            HelperClassDeriveRoot(node)

    def test_one_level(self):
        xml = '''
        <root>
            <register>
                <name>TimerCtrl0</name>
                <description>Timer Control Register</description>
                <addressOffset>0x0</addressOffset>
                <access>read-write</access>
                <resetValue>0x00008001</resetValue>
                <resetMask>0x0000ffff</resetMask>
                <size>32</size>
            </register>
            <register derivedFrom="TimerCtrl0">
                <name>TimerCtrl1</name>
                <description>Derived Timer</description>
                <addressOffset>0x4</addressOffset>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        test = HelperClassDeriveRoot(node)

        self.assertEqual(len(test.register), 2)

        self.assertEqual(type(test.register[0]), HelperClassDeriveRegister)
        self.assertEqual(test.register[0].name, "TimerCtrl0")
        self.assertEqual(test.register[0].description, "Timer Control Register")
        self.assertEqual(test.register[0].addressOffset, 0)
        self.assertEqual(test.register[0].size, 32)
        self.assertIsNone(test.register[0].derivedFrom)

        self.assertEqual(type(test.register[1]), HelperClassDeriveRegister)
        self.assertEqual(test.register[1].name, "TimerCtrl1")
        self.assertEqual(test.register[1].description, "Derived Timer")
        self.assertEqual(test.register[1].addressOffset, 4)
        self.assertEqual(test.register[1].size, 32)
        self.assertEqual(test.register[1].derivedFrom, test.register[0])

    def test_two_level(self):
        xml = '''
        <root>
            <register>
                <name>TimerCtrl0</name>
                <description>Timer Control Register</description>
                <addressOffset>0x0</addressOffset>
                <access>read-write</access>
                <resetValue>0x00008001</resetValue>
                <resetMask>0x0000ffff</resetMask>
                <size>32</size>
                <field>
                    <name>BitField0</name>
                    <description>Bit field 0</description>
                    <access>read-write</access>
                </field>
            </register>
            <register derivedFrom="TimerCtrl0">
                <name>TimerCtrl1</name>
                <description>Derived Timer</description>
                <addressOffset>0x4</addressOffset>
                <field derivedFrom="TimerCtrl0.BitField0">
                    <name>BitField1</name>
                    <description>Bit field 1</description>
                    <access>read-only</access>
                </field>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        test = HelperClassDeriveRoot(node)

        self.assertEqual(len(test.register), 2)

        self.assertEqual(type(test.register[0]), HelperClassDeriveRegister)
        self.assertEqual(test.register[0].name, "TimerCtrl0")
        self.assertEqual(test.register[0].description, "Timer Control Register")
        self.assertEqual(test.register[0].addressOffset, 0)
        self.assertEqual(test.register[0].size, 32)
        self.assertIsNone(test.register[0].derivedFrom)

        self.assertEqual(len(test.register[0].field), 1)

        self.assertEqual(test.register[0].field[0].name, "BitField0")
        self.assertEqual(test.register[0].field[0].description, "Bit field 0")
        self.assertEqual(test.register[0].field[0].access, pysvd.type.access.read_write)
        self.assertIsNone(test.register[0].field[0].derivedFrom)

        self.assertEqual(type(test.register[1]), HelperClassDeriveRegister)
        self.assertEqual(test.register[1].name, "TimerCtrl1")
        self.assertEqual(test.register[1].description, "Derived Timer")
        self.assertEqual(test.register[1].addressOffset, 4)
        self.assertEqual(test.register[1].size, 32)
        self.assertEqual(test.register[1].derivedFrom, test.register[0])

        self.assertEqual(len(test.register[1].field), 2)

        self.assertEqual(test.register[1].field[0].name, "BitField0")
        self.assertEqual(test.register[1].field[0].description, "Bit field 0")
        self.assertEqual(test.register[1].field[0].access, pysvd.type.access.read_write)
        self.assertIsNone(test.register[1].field[0].derivedFrom)

        self.assertEqual(test.register[1].field[1].name, "BitField1")
        self.assertEqual(test.register[1].field[1].description, "Bit field 1")
        self.assertEqual(test.register[1].field[1].access, pysvd.type.access.read_only)
        self.assertEqual(test.register[1].field[1].derivedFrom, test.register[0].field[0])

    def test_derive_from_derived(self):
        xml = '''
        <root>
            <register>
                <name>TimerCtrl0</name>
                <description>Timer Control Register</description>
                <addressOffset>0x0</addressOffset>
                <access>read-write</access>
                <resetValue>0x00008001</resetValue>
                <resetMask>0x0000ffff</resetMask>
                <size>32</size>
                <field>
                    <name>BitField0</name>
                    <description>Bit field 0</description>
                    <access>read-write</access>
                </field>
            </register>
            <register derivedFrom="TimerCtrl0">
                <name>TimerCtrl1</name>
                <description>Derived Timer</description>
                <addressOffset>0x4</addressOffset>
                <field derivedFrom="TimerCtrl0.BitField0">
                    <name>BitField1</name>
                    <description>Bit field 1</description>
                    <access>read-only</access>
                </field>
            </register>
            <register derivedFrom="TimerCtrl1">
                <name>TimerCtrl2</name>
                <description>Double Derived Timer</description>
                <addressOffset>0x8</addressOffset>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        test = HelperClassDeriveRoot(node)

        self.assertEqual(len(test.register), 3)

        self.assertEqual(type(test.register[0]), HelperClassDeriveRegister)
        self.assertEqual(test.register[0].name, "TimerCtrl0")
        self.assertEqual(test.register[0].description, "Timer Control Register")
        self.assertEqual(test.register[0].addressOffset, 0)
        self.assertEqual(test.register[0].size, 32)
        self.assertIsNone(test.register[0].derivedFrom)

        self.assertEqual(len(test.register[0].field), 1)

        self.assertEqual(test.register[0].field[0].name, "BitField0")
        self.assertEqual(test.register[0].field[0].description, "Bit field 0")
        self.assertEqual(test.register[0].field[0].access, pysvd.type.access.read_write)
        self.assertIsNone(test.register[0].field[0].derivedFrom)

        self.assertEqual(type(test.register[1]), HelperClassDeriveRegister)
        self.assertEqual(test.register[1].name, "TimerCtrl1")
        self.assertEqual(test.register[1].description, "Derived Timer")
        self.assertEqual(test.register[1].addressOffset, 4)
        self.assertEqual(test.register[1].size, 32)
        self.assertEqual(test.register[1].derivedFrom, test.register[0])

        self.assertEqual(len(test.register[1].field), 2)

        self.assertEqual(test.register[1].field[0].name, "BitField0")
        self.assertEqual(test.register[1].field[0].description, "Bit field 0")
        self.assertEqual(test.register[1].field[0].access, pysvd.type.access.read_write)
        self.assertIsNone(test.register[1].field[0].derivedFrom)

        self.assertEqual(test.register[1].field[1].name, "BitField1")
        self.assertEqual(test.register[1].field[1].description, "Bit field 1")
        self.assertEqual(test.register[1].field[1].access, pysvd.type.access.read_only)
        self.assertEqual(test.register[1].field[1].derivedFrom, test.register[0].field[0])

        self.assertEqual(type(test.register[2]), HelperClassDeriveRegister)
        self.assertEqual(test.register[2].name, "TimerCtrl2")
        self.assertEqual(test.register[2].description, "Double Derived Timer")
        self.assertEqual(test.register[2].addressOffset, 8)
        self.assertEqual(test.register[2].size, 32)
        self.assertEqual(test.register[2].derivedFrom, test.register[1])

        self.assertEqual(len(test.register[2].field), 2)

        self.assertEqual(test.register[2].field[0].name, "BitField0")
        self.assertEqual(test.register[2].field[0].description, "Bit field 0")
        self.assertEqual(test.register[2].field[0].access, pysvd.type.access.read_write)
        self.assertIsNone(test.register[2].field[0].derivedFrom)

        self.assertEqual(test.register[2].field[1].name, "BitField1")
        self.assertEqual(test.register[2].field[1].description, "Bit field 1")
        self.assertEqual(test.register[2].field[1].access, pysvd.type.access.read_only)
        self.assertEqual(test.register[2].field[1].derivedFrom, test.register[0].field[0])


class TestClassDim(unittest.TestCase):

    def test_inheritance(self):
        self.assertTrue(issubclass(pysvd.classes.Dim, pysvd.classes.Derive))
        self.assertTrue(issubclass(HelperClassDim, pysvd.classes.Dim))

    def test_no_name_exception(self):
        '''Required name field is missing'''

        xml = '''
        <root>
            <register>
                <dim>4</dim>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        with self.assertRaises(SyntaxError):
            pysvd.classes.Dim.add_elements(None, None, node, 'register')

    def test_no_dim(self):
        '''Normal generation if dim is missing'''

        xml = '''
        <root>
            <register>
                <name>Name</name>
                <description>Description</description>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        test = []
        pysvd.classes.Dim.add_elements(None, test, node, 'register')
        self.assertEqual(len(test), 1)

        self.assertEqual(test[0].name, "Name")
        self.assertEqual(test[0].description, "Description")

    def test_index_minimal(self):
        '''Minimal register generation'''

        xml = '''
        <root>
            <register>
                <dim>4</dim>
                <dimIncrement>2</dimIncrement>
                <name>PORT%s</name>
                <description>Port %s</description>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        test = []
        HelperClassDim.add_elements(None, test, node, 'register')
        self.assertEqual(len(test), 4)

        self.assertEqual(test[0].name, "PORT0")
        self.assertEqual(test[0].description, "Port 0")
        self.assertEqual(test[0].offset, 0)

        self.assertEqual(test[1].name, "PORT1")
        self.assertEqual(test[1].description, "Port 1")
        self.assertEqual(test[1].offset, 2)

        self.assertEqual(test[2].name, "PORT2")
        self.assertEqual(test[2].description, "Port 2")
        self.assertEqual(test[2].offset, 4)

        self.assertEqual(test[3].name, "PORT3")
        self.assertEqual(test[3].description, "Port 3")
        self.assertEqual(test[3].offset, 6)

    def test_index_fix(self):
        '''Fixed array generation'''

        xml = '''
        <root>
            <register>
                <dim>4</dim>
                <dimIncrement>4</dimIncrement>
                <name>MyArr[%s]</name>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        test = []
        pysvd.classes.Dim.add_elements(None, test, node, 'register')
        self.assertEqual(len(test), 1)

        self.assertEqual(test[0].name, "MyArr[4]")

    def test_index_list(self):
        '''Index generation by list'''

        xml = '''
        <root>
            <register>
                <dim>6</dim>
                <dimIncrement>4</dimIncrement>
                <dimIndex>A,B,C,D,E,Z</dimIndex>
                <name>GPIO_%s_CTRL</name>
                <description>GPIO Controller %s</description>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        test = []
        HelperClassDim.add_elements(None, test, node, 'register')
        self.assertEqual(len(test), 6)

        self.assertEqual(test[0].name, "GPIO_A_CTRL")
        self.assertEqual(test[0].description, "GPIO Controller A")
        self.assertEqual(test[0].offset, 0)
        self.assertEqual(test[1].name, "GPIO_B_CTRL")
        self.assertEqual(test[1].description, "GPIO Controller B")
        self.assertEqual(test[1].offset, 4)
        self.assertEqual(test[2].name, "GPIO_C_CTRL")
        self.assertEqual(test[2].description, "GPIO Controller C")
        self.assertEqual(test[2].offset, 8)
        self.assertEqual(test[3].name, "GPIO_D_CTRL")
        self.assertEqual(test[3].description, "GPIO Controller D")
        self.assertEqual(test[3].offset, 12)
        self.assertEqual(test[4].name, "GPIO_E_CTRL")
        self.assertEqual(test[4].description, "GPIO Controller E")
        self.assertEqual(test[4].offset, 16)
        self.assertEqual(test[5].name, "GPIO_Z_CTRL")
        self.assertEqual(test[5].description, "GPIO Controller Z")
        self.assertEqual(test[5].offset, 20)

    def test_index_range(self):
        '''Index generation by range'''

        xml = '''
        <root>
            <register>
                <dim>4</dim>
                <dimIncrement>8</dimIncrement>
                <dimIndex>3-6</dimIndex>
                <dimName>irq%s_t</dimName>
                <name>IRQ%s</name>
                <description>Interrupt %s</description>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        test = []
        HelperClassDim.add_elements(None, test, node, 'register')
        self.assertEqual(len(test), 4)

        self.assertEqual(test[0].name, "IRQ3")
        self.assertEqual(test[0].description, "Interrupt 3")
        self.assertEqual(test[0].dimName, "irq3_t")
        self.assertEqual(test[0].offset, 0)
        self.assertEqual(test[1].name, "IRQ4")
        self.assertEqual(test[1].description, "Interrupt 4")
        self.assertEqual(test[1].dimName, "irq4_t")
        self.assertEqual(test[1].offset, 8)
        self.assertEqual(test[2].name, "IRQ5")
        self.assertEqual(test[2].description, "Interrupt 5")
        self.assertEqual(test[2].dimName, "irq5_t")
        self.assertEqual(test[2].offset, 16)
        self.assertEqual(test[3].name, "IRQ6")
        self.assertEqual(test[3].description, "Interrupt 6")
        self.assertEqual(test[3].dimName, "irq6_t")
        self.assertEqual(test[3].offset, 24)

    def test_index_exception(self):
        '''dimIndex can not be interpreted as integer'''

        xml = '''
        <root>
            <register>
                <dim>4</dim>
                <dimIncrement>4</dimIncrement>
                <dimIndex>4x8</dimIndex>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        with self.assertRaises(ValueError):
            pysvd.classes.Dim.add_elements(None, None, node, 'register')

    def test_length_exception(self):
        '''dimIndex has less elements than dim requires'''

        xml = '''
        <root>
            <register>
                <dim>4</dim>
                <dimIncrement>4</dimIncrement>
                <dimIndex>A,B,C</dimIndex>
            </register>
        </root>'''

        node = ET.fromstring(xml)
        with self.assertRaises(AttributeError):
            pysvd.classes.Dim.add_elements(None, None, node, 'register')



import unittest
import pysvd


class TestTypeCpuName(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.cpuName.CM0PLUS), 'CM0PLUS')


class TestTypeEndian(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.endian.little), 'little')


class TestTypeDataType(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.dataType.uint32_t), 'uint32_t')


class TestTypeProtection(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.protection.non_secure), 'non-secure')


class TestTypeSauAccess(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.sauAccess.non_secure_callable_secure), 'non-secure-callable-secure')


class TestTypeAccess(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.access.read_write), 'read-write')


class TestTypeModifiedWriteValues(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.modifiedWriteValues.zeroToClear), 'zeroToClear')


class TestTypeReadAction(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.readAction.modify), 'modify')


class TestTypeEnumUsage(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.enumUsage.read_write), 'read-write')


class TestTypeAddressBlockUsage(unittest.TestCase):

    def test_string(self):
        self.assertEqual(str(pysvd.type.addressBlockUsage.registers), 'registers')



#!/usr/bin/env python3
# coding: utf-8
"""pysvd example project.

Read SVD file and generate C-style register access structs.
"""

import argparse
import xml.etree.ElementTree as ET
import pysvd
from enum import Enum

access = {
    pysvd.type.access.read_write: '__IO',
    pysvd.type.access.read_only: '__I',
    pysvd.type.access.write_only: '__O',
}

data_type = {
    8: 'uint8_t',
    16: 'uint16_t',
    32: 'uint32_t',
}

def format_registers(output, prefix, registers):
    data = []
    for register in registers:
        data.append(('`{1} <{0}.{1}_>`_'.format(prefix, register.name), '0x{:02X}'.format(register.addressOffset)))
    if len(data):
        output.write(table(('Register', 'Offset'), data))

    for register in registers:
        output.write('.. _{}.{}:\n\n'.format(prefix, register.name))

        output.write(underline(register.description, section.subsubsection))

        output.write(rst_list_name.format('Name', register.name))
        output.write(rst_list_name.format('Size', register.size))
        output.write(rst_list_name.format('Offset', '0x{:02X}'.format(register.addressOffset)))
        output.write(rst_list_name.format('Reset', '0x{:{fill}{width}X}'.format(register.resetValue, fill='0', width=register.size // 4)))
        output.write(rst_list_name.format('Access', register.access))
        output.write('\n')

        # Table

        # Bit description
        for field in register.fields:
            if field.bitWidth == 1:
                output.write('- Bit {} ({}) - {}\n'.format(field.bitOffset, field.access, field.name))
            else:
                output.write('- Bits {}:{} ({}) - {}\n'.format(
                    field.bitOffset + field.bitWidth - 1, field.bitOffset, field.access, field.name))
            output.write('   {}\n\n'.format(field.description))

            if hasattr(field, 'enumeratedValues'):
                for enumeratedValue in field.enumeratedValues.enumeratedValues:
                    output.write('   - {} - {}\n'.format(enumeratedValue.value, enumeratedValue.name))
                    if hasattr(enumeratedValue, 'description'):
                        output.write('      {}\n'.format(enumeratedValue.description))
                output.write('\n')


def write_register(register, output):
    output.write("    union\n    {\n")
    output.write("    {1} {2} {0.name};\n".format(register, access[register.access], data_type[register.size]))


def write_peripheral(peripheral, output):
    output.write("/* {0.name} - {0.description} */\n\n".format(peripheral))

    index = 1;
    offset = 0
    output.write("struct __attribute__((packed))  {0.name}_t\n{{\n".format(peripheral))
    for register in peripheral.registers:
        if register.addressOffset != offset:
            output.write("    __I {0} _{1};\n".format(data_type[register.addressOffset - offset]), index)
            index += 1
        write_register(register, output)
        offset = register.addressOffset + register.size // 8
    output.write("};\n")
    output.write("static_assert(sizeof({0}_t) == {1}, \"Size of {0}_t mismatch\");\n\n".format(peripheral.name, offset))

def main():
    parser = argparse.ArgumentParser(description='SVD to C-style register access structs')
    parser.add_argument('--svd', metavar='FILE', type=str, help='System view description (SVD) file', required=True)
    parser.add_argument('--output', '-o',  metavar='FILE', type=str, help='C output file', required=True)
    parser.add_argument('--version', action='version', version=pysvd.__version__)
    args = parser.parse_args()

    node = ET.parse(args.svd).getroot()
    device = pysvd.element.Device(node)

    output = open(args.output, "w")

    # Header
    output.write("/**\n" \
        " * @file\n" \
        " * @version {0.version}\n" \
        " * @brief Register access structs for {0.vendor} {0.name}\n" \
        " * @note This file is autogenerated using pysvd {1.__version__}\n" \
        " */\n\n".format(device, pysvd)
    )
    output.write("#pragma once\n\n")

    output.write("#define __I volatile const /*!< Read only permission */\n" \
        "#define __O volatile  /*!< Write only permission */\n" \
        "#define __IO volatile /*!< Read/write permission */\n\n"
    )

    # Interrupts
    interrupts = []
    for peripheral in device.peripherals:
        interrupts += peripheral.interrupts

    interrupt_count = 0
    interrupt_name = ""
    output.write("enum IRQn\n{\n")
    for interrupt in sorted(interrupts, key=lambda interrupt: interrupt.value):
        if interrupt_name != interrupt.name:
            output.write("    {0.name:20}= {0.value},\n".format(interrupt))
            interrupt_name = interrupt.name
            interrupt_count += 1
    output.write("};\n\n")

    # Peripherals
    data = []
    for peripheral in sorted(device.peripherals, key=lambda peripheral: peripheral.name):
        if peripheral.name.startswith('ADC'):
            write_peripheral(peripheral, output)
    output.close()

if __name__ == '__main__':
    main()



#!/usr/bin/python3
import argparse
import csv

def splitPort(name):
    return (name[1].lower(), int(name[2:]))

def namespace(port):
    return "using namespace {0}_type = ::controller::stm32::driver::gpio::{0}::type".format(port)

def input(port, index):
    return "template< gpio::type::pull_up_down_t pull_up_down_ >\n" \
        "using P{0}{1} = gpio::input< ::controller::GPIO{0}, {1}, pull_up_down_ >;".format(port.upper(), index)

def output(port, index):
    return "template< gpio::type::pull_up_down_t pull_up_down_,\n" \
        "    gpio::type::output_type_t output_type_,\n" \
        "    gpio::type::output_speed_t output_speed_>\n" \
        "using P{0}{1} = gpio::output< ::controller::GPIO{0}, {1}, pull_up_down_, output_type_, output_speed_ >;".format(port.upper(), index)

def analog(port, index):
    return "using P{0}{1} = gpio::analog< ::controller::GPIO{0}, {1} >;".format(port.upper(), index)

def alternate_function(port, index, af_num):
    return "template< gpio::type::pull_up_down_t pull_up_down_,\n" \
        "    gpio::type::output_type_t output_type_,\n" \
        "    gpio::type::output_speed_t output_speed_>\n" \
        "using P{0}{1} = gpio::alternate_function< ::controller::GPIO{0}, {1}, pull_up_down_, output_type_, output_speed_, gpio::type::alternate_function_t::af{2} >;".format(port.upper(), index, af_num)

if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(description="GPIO generator from CSV table.")
    arg_parser.add_argument('--csv', '-c', metavar='csvfile', type=str, required=True, help="CSV file path")
    args = arg_parser.parse_args()

    input_outputs = []
    analogs = []
    alternate_functions = []

    with open(args.csv, newline='') as csvfile:
        spamreader = csv.reader(csvfile)

        first_line = True
        for row in spamreader:
            if first_line:
                header = row
                first_line = False
            else:
                data = dict(zip(header, row))

                port, index = splitPort(data['Pin'])
                input_outputs.append({'port': port, 'index': index})

                add_func = 'Additional functions'
                if len(data[add_func]) != 0:
                    for name in data[add_func].split('/'):
                        analogs.append({'name': name, 'port': port, 'index': index})

                for af_num in range(8):
                    af = 'Alternate function {}'.format(af_num)
                    if len(data[af]) != 0:
                        for name in data[af].split('/'):
                            alternate_functions.append({'name': name, 'port': port, 'index': index, 'af_num': af_num})

    input_outputs.sort(key=lambda obj: (obj['port'], obj['index']))
    analogs.sort(key=lambda obj: (obj['name'], obj['port'], obj['index']))
    alternate_functions.sort(key=lambda obj: (obj['name'], obj['port'], obj['index'], obj['af_num']))
    #sorted(input_output, cmd)

    print('namespace input {\n')
    for io in input_outputs:
        print(input(io['port'], io['index']))
        print()
    print('}\n')

    print('namespace output {\n')
    for io in input_outputs:
        print(output(io['port'], io['index']))
        print()
    print('}\n')

    last_name = ''
    print('namespace analog {\n')
    for ana in analogs:
        if ana['name'] != last_name:
            if len(last_name):
                print('}\n')
            print('namespace {} {{\n'.format(ana['name']))
            last_name = ana['name']
        print(analog(ana['port'], ana['index']))
        print()
    print('}\n')
    print('}\n')

    last_name = ''
    print('namespace alternate_function {\n')
    for af in alternate_functions:
        if af['name'] != last_name:
            if len(last_name):
                print('}\n')
            print('namespace {} {{\n'.format(af['name']))
            last_name = af['name']
        print(alternate_function(af['port'], af['index'], af['af_num']))
        print()
    print('}\n')
    print('}\n')



"""SVD types from XSD schema file v1.3.3

Note: Type order taken from XSD file.
"""

from enum import Enum


class cpuName(Enum):
    """Specifies a selection of Cortex-M and Secure-Cores. This list will get
    extended as new processors are released.
    """

    CM0 = 'CM0'
    CM0PLUS = 'CM0PLUS'
    CM1 = 'CM1'
    SC000 = 'SC000'
    CM23 = 'CM23'
    CM3 = 'CM3'
    CM33 = 'CM33'
    CM35P = 'CM35P'
    SC300 = 'SC300'
    CM4 = 'CM4'
    CM7 = 'CM7'
    ARMV8MML = 'ARMV8MML'
    ARMV8MBL = 'ARMV8MBL'
    CA5 = 'CA5'
    CA7 = 'CA7'
    CA8 = 'CA8'
    CA9 = 'CA9'
    CA15 = 'CA15'
    CA17 = 'CA17'
    CA53 = 'CA53'
    CA57 = 'CA57'
    CA72 = 'CA72'
    other = 'other'

    def __str__(self):
        return self.value


# revision
# Specifies the CPU revision format as defined by ARM (rNpM)


class endian(Enum):
    """Pre-defines the tokens for specifying the endianess of the device"""

    little = 'little'
    big = 'big'
    selectable = 'selectable'
    other = 'other'

    def __str__(self):
        return self.value


class dataType(Enum):
    """Pre-defines the tokens in line with CMSIS data type definitions"""

    uint8_t = 'uint8_t'
    uint16_t = 'uint16_t'
    uint32_t = 'uint32_t'
    uint64_t = 'uint64_t'
    int8_t = 'int8_t'
    int16_t = 'int16_t'
    int32_t = 'int32_t'
    int64_t = 'int64_t'

    uint8_t_ptr = 'uint8_t *'
    uint16_t_ptr = 'uint16_t *'
    uint32_t_ptr = 'uint32_t *'
    uint64_t_ptr = 'uint64_t *'
    int8_t_ptr = 'int8_t *'
    int16_t_ptr = 'int16_t *'
    int32_t_ptr = 'int32_t *'
    int64_t_ptr = 'int64_t *'

    def __str__(self):
        return self.value


# nvicPrioBits
# Specifies the integer value range for the number of bits used in NVIC to
# encode priority levels


# dimableIdentifier
# Specifies the subset and sequence of characters used for specifying
# identifiers that may contain %s from dim. This is particularly important as
# these are used in ANSI C Structures during the device header file generation.


# identifier
# Specifies the subset and sequence of characters used for specifying
# identifiers that must not contain %s from dim. This is particularly
# important as these are used in ANSI C Structures during the device header
# file generation.


class protection(Enum):
    """V1.3: Protection Access Attribute Strings"""

    none = ''
    secure = 's'
    non_secure = 'n'
    privileged = 'p'

    def __str__(self):
        return self.name.replace('_', '-')


class sauAccess(Enum):
    """V1.3: SAU Access Type"""

    none = ''
    non_secure_callable_secure = 'c'
    non_secure = 'n'

    def __str__(self):
        return self.name.replace('_', '-')


# dimIndex
# Specifies the subset and sequence of characters used for specifying the
# sequence of indices in register arrays


# scaledNonNegativeInteger
# Specifies the format in which numbers are represented in hexadecimal or
# decimal format


# enumeratedValueData
# Specifies the number formats for the values in enumeratedValues


class access(Enum):
    """Specfies the pre-defined tokens for the available accesses"""

    read_only = 'read-only'
    write_only = 'write-only'
    read_write = 'read-write'
    writeOnce = 'writeOnce'
    read_writeOnce = 'read-writeOnce'

    def __str__(self):
        return self.value


class modifiedWriteValues(Enum):
    """Specifies the pre-defined tokens for the write side effects"""

    oneToClear = 'oneToClear'
    oneToSet = 'oneToSet'
    oneToToggle = 'oneToToggle'
    zeroToClear = 'zeroToClear'
    zeroToSet = 'zeroToSet'
    zeroToToggle = 'zeroToToggle'
    clear = 'clear'
    set = 'set'
    modify = 'modify'

    def __str__(self):
        return self.value


class readAction(Enum):
    """Specifies the pre-defined tokens for read side effects"""

    clear = 'clear'
    set = 'set'
    modify = 'modify'
    modifyExternal = 'modifyExternal'

    def __str__(self):
        return self.value


class enumUsage(Enum):
    """Specifies the pre-defined tokens for selecting what access types an
    enumeratedValues set is associated with
    """

    read = 'read'
    write = 'write'
    read_write = 'read-write'

    def __str__(self):
        return self.value


class addressBlockUsage(Enum):
    registers = 'registers'
    buffer = 'buffer'
    reserved = 'reserved'

    def __str__(self):
        return self.value



"""SVD elements from XSD schema file v1.3.3
"""
import re
import pysvd


def compare_attribute(lhs, rhs, attibute):
    """Compare attibute of objects.
    """
    if hasattr(lhs, attibute) != hasattr(rhs, attibute):
        return False

    if hasattr(lhs, attibute):
        return getattr(lhs, attibute) == getattr(rhs, attibute)

    return True

# /device
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_device.html
class Device(pysvd.classes.Base):
    """The element <device> provides the outermost frame of the description.

    * Only one <device> section is allowed per file. All other elements are described within this scope.
    * A <device> contains one or more peripherals, but one <cpu> description.
    * Optional elements such as <size>, <access>, or <resetValue> defined on this level represent default values for registers and can be
      refined at lower levels.
    """

    def __init__(self, node):
        self.peripherals = []

        super().__init__(node)

    def parse(self, node):
        super().parse(node)

        self.__dict__['schemaVersion'] = pysvd.parser.Text(pysvd.node.Attribute(node, 'schemaVersion', True))

        self.add_attribute(node, 'vendor', pysvd.parser.Text)
        self.add_attribute(node, 'vendorID', pysvd.parser.Text)
        self.add_attribute(node, 'name', pysvd.parser.Text, True)
        self.add_attribute(node, 'series', pysvd.parser.Text)
        self.add_attribute(node, 'version', pysvd.parser.Text, True)
        self.add_attribute(node, 'description', pysvd.parser.Text, True)
        self.add_attribute(node, 'licenseText', pysvd.parser.Text)
        self.add_attribute(node, 'headerSystemFilename', pysvd.parser.Text)
        self.add_attribute(node, 'headerDefinitionsPrefix', pysvd.parser.Text)
        self.add_attribute(node, 'addressUnitBits', pysvd.parser.Integer, True)
        self.add_attribute(node, 'width', pysvd.parser.Integer, True)

        # property group
        self.add_attribute(node, 'size', pysvd.parser.Integer, False, 32)
        self.add_enum_attribute(node, 'access', pysvd.type.access, False, pysvd.type.access.read_write)
        self.add_enum_attribute(node, 'protection', pysvd.type.protection, False, pysvd.type.protection.none)
        self.add_attribute(node, 'resetValue', pysvd.parser.Integer, False, 0x00000000)
        self.add_attribute(node, 'resetMask', pysvd.parser.Integer, False, 0xFFFFFFFF)

        # Clean up license text from whitespaces
        result = ''
        if hasattr(self, 'licenseText'):
            for line in self.licenseText.splitlines():
                line = line.strip()
                if len(line):
                    result += line + '\n'
            self.licenseText = result

        cpu_node = node.find('cpu')
        if cpu_node is not None:
            self.cpu = pysvd.element.Cpu(self, cpu_node)

        peripherals_node = node.find('./peripherals')
        if peripherals_node is None:
            raise SyntaxError("No element 'peripherals' found in 'device'")

        Peripheral.add_elements(self, self.peripherals, peripherals_node, 'peripheral')
        if len(self.peripherals) < 1:
            raise SyntaxError("At least one element of 'peripheral' is mandatory in 'peripherals'")

    def find(self, name):
        """Find peripheral by name."""
        for peripheral in self.peripherals:
            if peripheral.name == name:
                return peripheral
        return None


# /device/cpu
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_cpu.html
class Cpu(pysvd.classes.Parent):
    """The CPU section describes the processor included in the microcontroller device. This section is mandatory if the SVD file is used to
    generate the device header file.
    """

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        self.__dict__['name'] = pysvd.parser.Enum(pysvd.type.cpuName, pysvd.node.Element(node, 'name', True).replace('+', 'PLUS'))
        self.add_attribute(node, 'revision', pysvd.parser.Text, True)
        self.add_enum_attribute(node, 'endian', pysvd.type.endian, True)
        self.add_attribute(node, 'mpuPresent', pysvd.parser.Boolean, True)
        self.add_attribute(node, 'fpuPresent', pysvd.parser.Boolean, True)
        self.add_attribute(node, 'fpuDP', pysvd.parser.Boolean)
        self.add_attribute(node, 'dspPresent', pysvd.parser.Boolean)
        self.add_attribute(node, 'icachePresent', pysvd.parser.Boolean)
        self.add_attribute(node, 'dcachePresent', pysvd.parser.Boolean)
        self.add_attribute(node, 'itcmPresent', pysvd.parser.Boolean)
        self.add_attribute(node, 'dtcmPresent', pysvd.parser.Boolean)
        self.add_attribute(node, 'vtorPresent', pysvd.parser.Boolean, False, True)

        self.add_attribute(node, 'nvicPrioBits', pysvd.parser.Integer, True)
        self.add_attribute(node, 'vendorSystickConfig', pysvd.parser.Boolean, True)

        self.add_attribute(node, 'deviceNumInterrupts', pysvd.parser.Integer)
        self.add_attribute(node, 'sauNumRegions', pysvd.parser.Integer)

        sau_regions_config_node = node.find('./sauRegionsConfig')
        if sau_regions_config_node is not None:
            self.sauRegionsConfig = SauRegionConfig(self, sau_regions_config_node)


# /device/cpu/sauRegionsConfig
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_cpu.html#elem_sauRegionsConfig
class SauRegionConfig(pysvd.classes.Group):
    """Set the configuration for the Secure Attribution Unit (SAU) when they are preconfigured by HW or Firmware.
    """

    attributes = ['protectionWhenDisabled']

    def __init__(self, parent, node):
        self.regions = []

        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        enabled = pysvd.parser.Boolean(pysvd.node.Attribute(node, 'enabled'))
        if enabled is not None:
            self.__dict__['enabled'] = enabled

        protectionWhenDisabled = pysvd.parser.Enum(pysvd.type.protection, pysvd.node.Attribute(node, 'protectionWhenDisabled'))
        if protectionWhenDisabled is not None:
            self.__dict__['protectionWhenDisabled'] = protectionWhenDisabled

        for sau_regions_config_region_node in node.findall('region'):
            self.regions.append(SauRegionsConfigRegion(self, sau_regions_config_region_node))


# /device/cpu/sauRegionsConfig/region
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_cpu.html#elem_region
class SauRegionsConfigRegion(pysvd.classes.Parent):
    """Define the regions of the Secure Attribution Unit (SAU). The protection level is inherited from the attribute
    <protectionWhenDisabled> of the enclosing element sauRegionsConfig.
    """

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        self.__dict__['enabled'] = pysvd.parser.Boolean(pysvd.node.Attribute(node, 'enabled'), True)
        name = pysvd.parser.Text(pysvd.node.Attribute(node, 'name'))
        if name is not None:
            self.__dict__['name'] = name

        self.add_attribute(node, 'base', pysvd.parser.Integer, True)
        self.add_attribute(node, 'limit', pysvd.parser.Integer, True)
        self.add_enum_attribute(node, 'access', pysvd.type.sauAccess, True)


# /device/peripherals
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_peripherals.html
# All peripherals of a device are enclosed within the tag <peripherals>.
#
# Note: This level is skipped for better readability

# /device/peripherals/peripheral
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_peripherals.html#elem_peripheral
class Peripheral(pysvd.classes.Dim):
    """At least one peripheral has to be defined.

    * Each peripheral describes all registers belonging to that peripheral.
    * The address range allocated by a peripheral is defined through one or more address blocks.
    * An address block and register addresses are specified relative to the base address of a peripheral. The address block information can
      be used for constructing a memory map for the device peripherals.

    Starting version 1.3 of the SVD specification, arrays of peripherals can be specified. The single peripheral description gets
    duplicated automatically into an array. The number of array elements is specified by the <dim> element. To define arrays, the <name>
    needs the format myPeripheral[%s]. The tag <dimIncrement> specifies the address offset between two peripherals. To create copies of a
    peripheral using different names, you must use the derivedFrom attribute.
    """

    def __init__(self, parent, node):
        self.addressBlocks = []
        self.registers = []
        self.clusters = []

        super().__init__(parent, node)

    def __eq__(self, other):
        """Compare element and all attributes.
        """
        if not isinstance(other, Peripheral):
            return NotImplemented

        if not self.equal_struct(other):
            return False

        return compare_attribute(self, other, 'name') and \
            compare_attribute(self, other, 'description') and \
            compare_attribute(self, other, 'version') and \
            compare_attribute(self, other, 'alternatePeripheral') and \
            compare_attribute(self, other, 'groupName') and \
            compare_attribute(self, other, 'prependToName') and \
            compare_attribute(self, other, 'appendToName') and \
            compare_attribute(self, other, 'headerStructName') and \
            compare_attribute(self, other, 'disableCondition') and \
            compare_attribute(self, other, 'baseAddress') and \
            compare_attribute(self, other, 'addressBlock') and \
            compare_attribute(self, other, 'interrupt')

    def equal_struct(self, other):
        """Compare only child elements.
        """
        if not isinstance(other, Peripheral):
            return NotImplemented

        if len(self.registers) != len(other.registers):
            return False

        for (lhs, rhs) in zip(self.registers, other.registers):
            if lhs != rhs:
                return False

        if len(self.clusters) != len(other.clusters):
            return False

        for (lhs, rhs) in zip(self.clusters, other.clusters):
            if lhs != rhs:
                return False
        return True

    def set_offset(self, value):
        self.baseAddress += value

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'version', pysvd.parser.Text)
        self.add_attribute(node, 'alternatePeripheral', pysvd.parser.Text)
        self.add_attribute(node, 'groupName', pysvd.parser.Text)
        self.add_attribute(node, 'prependToName', pysvd.parser.Text)
        self.add_attribute(node, 'appendToName', pysvd.parser.Text)
        self.add_attribute(node, 'headerStructName', pysvd.parser.Text)
        self.add_attribute(node, 'disableCondition', pysvd.parser.Text)
        self.add_attribute(node, 'baseAddress', pysvd.parser.Integer, True)

        AddressBlock.add_elements(self, self.addressBlocks, node, 'addressBlock')

        self.interrupts = []
        Interrupt.add_elements(self, self.interrupts, node, 'interrupt')

        registers_node = node.find('./registers')
        if registers_node is not None:
            Register.add_elements(self, self.registers, registers_node, 'register')
            Cluster.add_elements(self, self.clusters, registers_node, 'cluster')

            if len(self.registers) < 1 and len(self.clusters) < 1:
                raise SyntaxError("At least one element of 'register' or 'cluster' is mandatory in 'registers'")

    def find(self, name):
        """Find cluster and register by name."""
        for cluster in self.clusters:
            if cluster.name == name:
                return cluster

        for register in self.registers:
            if register.name == name:
                return register
        return None


# /device/peripherals/peripheral/addressBlock
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_peripherals.html#elem_addressBlock
class AddressBlock(pysvd.classes.Group):
    """Specify an address range uniquely mapped to this peripheral. A peripheral must have at least one address block. If a peripheral is
    derived form another peripheral, the <addressBlock> is not mandatory.
    """

    attributes = ['protection']

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'offset', pysvd.parser.Integer, True)
        self.add_attribute(node, 'size', pysvd.parser.Integer, True)
        self.add_enum_attribute(node, 'usage', pysvd.type.addressBlockUsage, True)
        self.add_enum_attribute(node, 'protection', pysvd.type.protection)


# /device/peripherals/peripheral/interrupt
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_peripherals.html#elem_interrupt
class Interrupt(pysvd.classes.Parent):
    """A peripheral can have multiple interrupts. This entry allows the debugger to show interrupt names instead of interrupt numbers.
    """

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'name', pysvd.parser.Text, True)
        self.add_attribute(node, 'description', pysvd.parser.Text)
        self.add_attribute(node, 'value', pysvd.parser.Integer, True)


# /device/peripherals/peripheral/registers
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_registers.html
# All registers of a peripheral are enclosed between the <registers> opening and closing tags. Clusters define a set of registers. You can
# either use the <cluster> or the <register> element.
#
# Note: This level is skipped for better readability


# /device/pripherals/peripheral/registers/.../cluster
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_registers.html#elem_cluster
class Cluster(pysvd.classes.Dim):
    """Cluster describes a sequence of neighboring registers within a peripheral. A <cluster> specifies the addressOffset relative to the
    baseAddress of the grouping element. All <register> elements within a <cluster> specify their addressOffset relative to the cluster
    base address (<peripheral.baseAddress> + <cluster.addressOffset>).

    Multiple <register> and <cluster> sections may occur in any order. Since version 1.3 of the specification, the nesting of <cluster>
    elements is supported. Nested clusters express hierarchical structures of registers. It is predominantely targeted at the generation of
    device header files to create a C-data structure within the peripheral structure instead of a flat list of registers. Note, you can
    also specify an array of a cluster using the <dim> element.
    """

    def __init__(self, parent, node):
        self.registers = []
        self.clusters = []

        super().__init__(parent, node)

    def set_offset(self, value):
        self.addressOffset += value

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'name', pysvd.parser.Text, True)
        self.add_attribute(node, 'description', pysvd.parser.Text, self.derivedFrom is not None)
        self.add_attribute(node, 'alternateCluster', pysvd.parser.Text)
        self.add_attribute(node, 'headerStructName', pysvd.parser.Text)
        self.add_attribute(node, 'addressOffset', pysvd.parser.Integer, True)

        self.add_attribute(node, 'size', pysvd.parser.Integer)
        self.add_enum_attribute(node, 'access', pysvd.type.access)
        self.add_enum_attribute(node, 'protection', pysvd.type.protection)
        self.add_attribute(node, 'resetValue', pysvd.parser.Integer)
        self.add_attribute(node, 'resetMask', pysvd.parser.Integer)

        Register.add_elements(self, self.registers, node, 'register')
        Cluster.add_elements(self, self.clusters, node, 'cluster')

    def find(self, name):
        """Find cluster and register by name."""
        for cluster in self.clusters:
            if cluster.name == name:
                return cluster

        for register in self.registers:
            if register.name == name:
                return register
        return None


# /device/peripherals/peripheral/registers/.../register
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_registers.html#elem_register
class Register(pysvd.classes.Dim):
    """The description of registers is the most essential part of SVD. If the elements <size>, <access>, <resetValue>, and <resetMask>
    have not been specified on a higher level, then these elements are mandatory on register level.

    A register can represent a single value or can be subdivided into individual bit-fields of specific functionality and semantics. From
    a schema perspective, the element <fields> is optional, however, from a specification perspective, <fields> are mandatory when they are
    described in the device documentation.

    You can define register arrays where the single description gets duplicated automatically. The size of the array is specified by the
    <dim> element. Register names get composed by the element <name> and the index-specific string defined in <dimIndex>. The element
    <dimIncrement> specifies the address offset between two registers.
    """

    def __init__(self, parent, node):
        self.fields = []

        super().__init__(parent, node)

    def __eq__(self, other):
        """Compare element and all attributes.
        """
        if not isinstance(other, Register):
            return NotImplemented

        if not self.equal_struct(other):
            return False

        return compare_attribute(self, other, 'name') and \
            compare_attribute(self, other, 'displayName') and \
            compare_attribute(self, other, 'description') and \
            compare_attribute(self, other, 'alternateGroup') and \
            compare_attribute(self, other, 'alternateRegister') and \
            compare_attribute(self, other, 'addressOffset') and \
            compare_attribute(self, other, 'size') and \
            compare_attribute(self, other, 'access') and \
            compare_attribute(self, other, 'protection') and \
            compare_attribute(self, other, 'resetValue') and \
            compare_attribute(self, other, 'resetMask') and \
            compare_attribute(self, other, 'dataType') and \
            compare_attribute(self, other, 'modifiedWriteValues') and \
            compare_attribute(self, other, 'readAction')

    def equal_struct(self, other):
        """Compare only child elements.
        """
        if not isinstance(other, Register):
            return NotImplemented

        if len(self.fields) != len(other.fields):
            return False

        for (lhs, rhs) in zip(self.fields, other.fields):
            if lhs != rhs:
                return False
        return True

    def set_offset(self, value):
        self.addressOffset += value

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'name', pysvd.parser.Text, True)
        self.add_attribute(node, 'displayName', pysvd.parser.Text)
        self.add_attribute(node, 'description', pysvd.parser.Text, self.derivedFrom is not None)
        self.add_attribute(node, 'alternateGroup', pysvd.parser.Text)
        self.add_attribute(node, 'alternateRegister', pysvd.parser.Text)
        self.add_attribute(node, 'addressOffset', pysvd.parser.Integer, True)

        self.add_attribute(node, 'size', pysvd.parser.Integer)
        self.add_enum_attribute(node, 'access', pysvd.type.access)
        self.add_enum_attribute(node, 'protection', pysvd.type.protection)
        self.add_attribute(node, 'resetValue', pysvd.parser.Integer)
        self.add_attribute(node, 'resetMask', pysvd.parser.Integer)

        self.add_enum_attribute(node, 'dataType', pysvd.type.dataType)
        self.add_enum_attribute(node, 'modifiedWriteValues', pysvd.type.modifiedWriteValues, False, pysvd.type.modifiedWriteValues.modify)
        self.add_enum_attribute(node, 'readAction', pysvd.type.readAction)

        write_constraint_node = node.find('./writeConstraint')
        if write_constraint_node is not None:
            self.writeConstraint = WriteConstraint(self, write_constraint_node)

        fields_node = node.find('./fields')
        if fields_node is not None:
            Field.add_elements(self, self.fields, fields_node, 'field')

            if len(self.fields) < 1:
                raise SyntaxError("At least one element of 'field' is mandatory in 'fields'")

    def find(self, name):
        """Find field by name."""
        for field in self.fields:
            if field.name == name:
                return field
        return None


# /device/peripherals/peripheral/registers/.../register/.../writeConstraint
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_registers.html#elem_writeConstraint
class WriteConstraint(pysvd.classes.Parent):
    """Define constraints for writing values to a field. You can choose between three options, which are mutualy exclusive.
    """

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        writeAsRead = pysvd.node.Element(node, 'writeAsRead')
        useEnumeratedValues = pysvd.node.Element(node, 'useEnumeratedValues')
        range_node = node.find("./range")
        if writeAsRead is not None:
            self.__dict__['writeAsRead'] = pysvd.parser.Boolean(writeAsRead)
        elif useEnumeratedValues is not None:
            self.__dict__['useEnumeratedValues'] = pysvd.parser.Boolean(useEnumeratedValues)
        else:
            rangeMinimum = None
            rangeMaximum = None
            if range_node is not None:
                rangeMinimum = pysvd.node.Element(range_node, 'minimum')
                rangeMaximum = pysvd.node.Element(range_node, 'maximum')

            if rangeMinimum is None or rangeMaximum is None:
                raise SyntaxError("Either 'writeAsRead', 'useEnumeratedValues' or 'range' is mandatory in 'writeConstraint'")

            self.__dict__['rangeMinimum'] = pysvd.parser.Integer(rangeMinimum)
            self.__dict__['rangeMaximum'] = pysvd.parser.Integer(rangeMaximum)


# /device/peripherals/peripheral/registers/.../register/fields
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_registers.html#elem_fields
# Grouping element to define bit-field properties of a register.
#
# Note: This level is skipped for better readability


# /device/peripherals/peripheral/registers/.../fields/field
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_registers.html#elem_field
class Field(pysvd.classes.Dim):
    """All fields of a register are enclosed between the <fields> opening and closing tags.

    A bit-field has a name that is unique within the register. The position and size within the register can be decsribed in two ways:
    * by the combination of the least significant bit's position (lsb) and the most significant bit's position (msb), or
    * the lsb and the bit-width of the field.

    A field may define an enumeratedValue in order to make the display more intuitive to read.
    """

    attributes = ['access']

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def __eq__(self, other):
        if not isinstance(other, Field):
            return NotImplemented

        if not self.equal_struct(other):
            return False

        return compare_attribute(self, other, 'name') and \
            compare_attribute(self, other, 'description') and \
            compare_attribute(self, other, 'bitOffset') and \
            compare_attribute(self, other, 'bitWidth') and \
            compare_attribute(self, other, 'access') and \
            compare_attribute(self, other, 'modifiedWriteValues') and \
            compare_attribute(self, other, 'readAction') and \
            compare_attribute(self, other, 'writeConstraint')

    def equal_struct(self, other):
        """Check structure elements, but not overridable attributes.
        """
        if not isinstance(other, Field):
            return NotImplemented

        return compare_attribute(self, other, 'enumeratedValues')

        x = """
        if len(self.enumeratedValues) != len(other.enumeratedValues):
            return False

        for (lhs, rhs) in zip(self.enumeratedValues, other.enumeratedValues):
            if lhs != rhs:
                return False
        return True
        """


    def set_offset(self, value):
        self.bitOffset += value

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'name', pysvd.parser.Text, True)
        self.add_attribute(node, 'description', pysvd.parser.Text, self.derivedFrom is not None)

        # bitRangeOffsetWidthStyle
        bitOffset = pysvd.parser.Integer(pysvd.node.Element(node, 'bitOffset'))
        bitWidth = pysvd.parser.Integer(pysvd.node.Element(node, 'bitWidth'))
        if bitOffset is not None:
            # If bitWidth is not set, default is 1
            bitWidth = 1 if bitWidth is None else bitWidth
        else:
            # bitRangeLsbMsbStyle
            lsb = pysvd.parser.Integer(pysvd.node.Element(node, 'lsb'))
            msb = pysvd.parser.Integer(pysvd.node.Element(node, 'msb'))
            if lsb is None or msb is None:
                bitRange = pysvd.parser.Text(pysvd.node.Element(node, 'bitRange'))
                if bitRange is None:
                    raise ValueError("Field '{}' has no valid bit-range".format(self.name if hasattr(self, 'name') else '<unknown>'))

                match = re.search(r'\[([0-9]+):([0-9]+)\]', bitRange)
                lsb = int(match.group(2))
                msb = int(match.group(1))

            bitOffset = lsb
            bitWidth = (msb - lsb) + 1

        self.__dict__['bitOffset'] = bitOffset
        self.__dict__['bitWidth'] = bitWidth

        self.add_enum_attribute(node, 'access', pysvd.type.access)
        self.add_enum_attribute(node, 'modifiedWriteValues', pysvd.type.modifiedWriteValues, False, pysvd.type.modifiedWriteValues.modify)
        self.add_enum_attribute(node, 'readAction', pysvd.type.readAction)

        write_constraint_node = node.find('./writeConstraint')
        if write_constraint_node is not None:
            self.writeConstraint = WriteConstraint(self, write_constraint_node)

        enumerated_values_node = node.find('./enumeratedValues')
        if enumerated_values_node is not None:
            self.enumeratedValues = EnumeratedValues(self, enumerated_values_node)

    def find(self, name):
        """Find enumeratedValues by name."""
        if hasattr(self, 'enumeratedValues') and hasattr(self.enumeratedValues, 'name') and self.enumeratedValues.name == name:
            return self.enumeratedValues
        return None


# /device/peripherals/peripheral/registers/.../field/enumeratedValues
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_registers.html#elem_enumeratedValues
class EnumeratedValues(pysvd.classes.Derive):
    """The concept of enumerated values creates a map between unsigned integers and an identifier string. In addition, a description string
    can be associated with each entry in the map.

    0 <-> disabled -> "The clock source clk0 is turned off."
    1 <-> enabled  -> "The clock source clk1 is running."
    2 <-> reserved -> "Reserved values. Do not use."
    3 <-> reserved -> "Reserved values. Do not use."

    This information generates an enum in the device header file. The debugger may use this information to display the identifier string as
    well as the description. Just like symbolic constants making source code more readable, the system view in the debugger becomes more
    instructive. The detailed description can provide reference manual level details within the debugger.
    """

    def __init__(self, parent, node):
        self.enumeratedValues = []

        super().__init__(parent, node)

    def __eq__(self, other):
        if not isinstance(other, EnumeratedValues):
            return NotImplemented

        if not self.equal_struct(other):
            return False

        return compare_attribute(self, other, 'name') and \
            compare_attribute(self, other, 'headerEnumName') and \
            compare_attribute(self, other, 'usage')

    def equal_struct(self, other):
        if not isinstance(other, EnumeratedValues):
            return NotImplemented

        if len(self.enumeratedValues) != len(other.enumeratedValues):
            return False

        for (lhs, rhs) in zip(self.enumeratedValues, other.enumeratedValues):
            if lhs != rhs:
                return False
        return True

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'name', pysvd.parser.Text)
        self.add_attribute(node, 'headerEnumName', pysvd.parser.Text)
        self.add_enum_attribute(node, 'usage', pysvd.type.enumUsage, False, pysvd.type.enumUsage.read_write)

        for child in node.findall('./enumeratedValue'):
            self.enumeratedValues.append(EnumeratedValue(self, child))

        if len(self.enumeratedValues) < 1:
            raise SyntaxError("At least one element of enumeratedValue is needed in enumeratedValues '{}'".format(
                self.name if hasattr(self, 'name') else '<unknown>'))


# /device/peripherals/peripheral/registers/.../enumeratedValue
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_registers.html#elem_enumeratedValue
class EnumeratedValue(pysvd.classes.Parent):
    """An enumeratedValue defines a map between an unsigned integer and a string.
    """

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def __eq__(self, other):
        if not isinstance(other, EnumeratedValue):
            return NotImplemented

        return compare_attribute(self, other, 'name') and \
            compare_attribute(self, other, 'value') and \
            compare_attribute(self, other, 'description') and \
            compare_attribute(self, other, 'isDefault')

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'name', pysvd.parser.Text)
        self.add_attribute(node, 'description', pysvd.parser.Text)
        self.add_attribute(node, 'value', pysvd.parser.Integer)
        self.add_attribute(node, 'isDefault', pysvd.parser.Boolean)

        if not hasattr(self, 'value') and not hasattr(self, 'isDefault'):
            raise SyntaxError("Either 'value' or 'isDefault' is mandatory in enumeratedValue '{}'".format(
                self.name if hasattr(self, 'name') else '<unknown>'))


# /device/peripherals/peripheral/.../dimArrayIndex
# http://www.keil.com/pack/doc/cmsis/svd/html/elem_special.html#elem_dimArrayIndex
class DimArrayIndex(pysvd.classes.Parent):
    """This information is used for generating an enum in the device header file. The debugger may use this information to display the
    identifier string as well as the description. Just like symbolic constants making source code more readable, the system view in the
    debugger becomes more instructive.
    """

    def __init__(self, parent, node):
        self.enumeratedValues = []

        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        self.add_attribute(node, 'headerEnumName', pysvd.parser.Text)

        for child in node.findall('./enumeratedValue'):
            self.enumeratedValues.append(EnumeratedValue(self, child))

        if len(self.enumeratedValues) < 1:
            raise SyntaxError("At least one element of enumeratedValue is needed in dimArrayIndex '{}'".format(
                self.headerEnumNames if hasattr(self, 'headerEnumName') else '<unknown>'))



#!/usr/bin/env python3
# coding: utf-8
"""Read SVD file, order elements, check for valid elements to generate register access structs and displays possible substitutions.
"""

import sys
import argparse
import xml.etree.ElementTree as ET
from enum import IntEnum
import itertools
from natsort import natsorted
from colorama import Fore, Back, Style

import pysvd

class Level(IntEnum):
    """Output level."""
    warning = 0
    hint = 1
    all = 2

class Depth(IntEnum):
    """Analysis depth."""
    peripherals = 0
    registers = 1
    fields = 2
    enumeratedValues = 3

def integer(value):
    """Convert binary, hex, octal and decimal strings to integer."""
    if value.startswith('#'):
        value = value.replace('#', '0b')

    if value.startswith('0b'):
        value = value.replace('x', '0')
    return int(value, 0)

def compare_peripherals(level, peripherals):
    """Compare peripherals."""

    print("{0.CYAN}Peripherals{0.RESET}".format(Fore))

    peripherals_base = []
    peripherals_derived = []
    peripherals_not_derived = []
    for a, b in itertools.combinations(peripherals, 2):
        # Check if item is already in list
        if a in peripherals_derived + peripherals_not_derived:
            continue

        if a.equal_struct(b):
            if a not in peripherals_base:
                peripherals_base.append(a)

            # Already derived peripheral
            if b.derivedFrom:
                if level >= Level.all:
                    print("[{0.GREEN}OK{0.RESET}] Peripheral '{2.name}' is derived from '{1.name}'".format(Fore, b.derivedFrom, b))
                peripherals_derived.append(b)
            # Derivable peripheral
            else:
                if level >= Level.warning:
                    print("[{0.RED}WARNING{0.RESET}] Peripheral '{2.name}' can be derived from '{1.name}'".format(Fore, a, b))
                peripherals_not_derived.append(b)

    peripherals_none_derivable = [peripheral for peripheral in peripherals \
        if peripheral not in peripherals_base and \
            peripheral not in peripherals_derived and \
            peripheral not in peripherals_not_derived]

    if level >= Level.hint:
        for peripheral in peripherals_none_derivable:
            print("[{0.YELLOW}HINT{0.RESET}] Peripheral '{1.name}' can not be derived".format(Fore, peripheral))
    print()

    return (peripherals_base, peripherals_none_derivable)

def compare_registers(level, peripheral, registers):
    print("{0.CYAN}Registers of peripheral '{1}'{0.RESET}".format(Fore, peripheral.name))

    registers_base = []
    registers_derived = []
    registers_not_derived = []

    for a, b in itertools.combinations(registers, 2):
        # Check if item is already in list
        if a in registers_derived + registers_not_derived:
            continue

        if a.equal_struct(b):
            if a not in registers_base:
                registers_base.append(a)

            # Already derived register
            if b.derivedFrom:
                if level >= Level.all:
                    print("[{0.GREEN}OK{0.RESET}] Register '{2.name}' is derived from '{1.name}'".format(Fore, b.derivedFrom, b))
                registers_derived.append(b)
            # Derivable register
            else:
                if level >= Level.warning:
                    print("[{0.RED}WARNING{0.RESET}] Register '{2.name}' can be derived from '{1.name}'".format(Fore, a, b))
                registers_not_derived.append(b)

    registers_none_derivable = [register for register in registers \
        if register not in registers_base and \
            register not in registers_derived and \
            register not in registers_not_derived]

    if level >= Level.hint:
        for register in registers_none_derivable:
            print("[{0.YELLOW}HINT{0.RESET}] Register '{1.name}' can not be derived".format(Fore, register))
    print()

    return (registers_base, registers_none_derivable)

def main():
    parser = argparse.ArgumentParser(description='Read SVD file, order elements, check for' \
        'valid elements to generate register access structs and displays possible substitutions.')
    parser.add_argument('--svd', metavar='FILE', type=str, help='System view description (SVD) file', required=True)
    parser.add_argument('--output', '-o', metavar='FILE', type=str, help='Save ordered SVD output file')
    parser.add_argument('--level', '-l', choices=['all', 'hint', 'warning'], help='Select level of output messages', default='all')
    parser.add_argument('--depth', '-d', choices=['peripherals', 'registers', 'fields', 'enumeratedValues'], help='Select depth of analysis', default='enumeratedValues')
    parser.add_argument('--sort', action='store_true', help='Sort elements before comparing')
    args = parser.parse_args()
    level = Level[args.level]
    depth = Depth[args.depth]

    xml = ET.parse(args.svd)

    if args.sort:
      print('Sort peripherals by name')
      for peripherals in xml.findall('.//peripherals'):
          peripherals[:] = natsorted(peripherals, key=lambda x: x.find('name').text)

      print('Sort registers by addressOffset')
      for registers in xml.findall('.//registers'):
          registers[:] = sorted(registers, key=lambda x: integer(x.find('addressOffset').text))

      print('Sort fields by bitOffset')
      for fields in xml.findall('.//fields'):
          fields[:] = sorted(fields, key=lambda x: integer(x.find('bitOffset').text))

      print('Sort enumeratedValues by value')
      for enumeratedValues in xml.findall('.//enumeratedValues'):
          enumeratedValues[:] = sorted(enumeratedValues, key=lambda x: integer(x.find('value').text) if x.tag == 'enumeratedValue' else -1)

    print('Remove linebreaks from description tags')
    for item in xml.findall('.//description'):
        text = item.text.replace('\n', ' ').strip()
        item.text = ' '.join(text.split())

    if args.output:
        xml.write(args.output, encoding="utf-8", xml_declaration=True, method="xml", short_empty_elements=True)
    print()

    # Load cleaned-up SVD file
    try:
        device = pysvd.element.Device(xml.getroot())
    except Exception as e:
        print("Error parsing SVD file: {}".format(str(e)))
        sys.exit(2)

    (peripherals_base, peripherals_none_derivable) = compare_peripherals(level, device.peripherals)
    if depth >= Depth.registers:
        peripherals = natsorted(peripherals_base + peripherals_none_derivable, key=lambda peripheral: peripheral.name)
        for peripheral in peripherals:
            compare_registers(level, peripheral, peripheral.registers)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# coding: utf-8
"""pysvd example project.

Read SVD file and generate device datasheet in rst format.
"""

import argparse
import xml.etree.ElementTree as ET
import pysvd
from enum import Enum

rst_list_name = ":{}: {}\n"


class section(Enum):
    section = '='
    subsection = '-'
    subsubsection = '^'
    paragraphs = '"'

    def __str__(self):
        return self.value


def underline(value, section):
    return "{}\n{}\n\n".format(value, str(section) * len(value))


def table(header, data):
    if len(data) == 0:
        raise ValueError("data has no length")

    if len(header) != len(data[0]):
        raise ValueError("columns in header and data do not match")

    # Find width of column
    widths = []
    for caption in header:
        widths.append(len(caption))

    for row in data:
        index = 0
        for text in row:
            widths[index] = max(widths[index], len(text))
            index += 1

    result = ''
    separator = []
    line = []

    index = 0
    for caption in header:
        separator.append('=' * widths[index])
        line.append('{:{align}{width}}'.format(caption, align='^', width=widths[index]))
        index += 1

    result = ' '.join(separator) + '\n'
    result += ' '.join(line).rstrip() + '\n'
    result += ' '.join(separator) + '\n'

    for row in data:
        index = 0
        for text in row:
            line[index] = '{:{width}}'.format(text, width=widths[index])
            index += 1
        result += ' '.join(line).rstrip() + '\n'
    result += ' '.join(separator) + '\n'

    return result + '\n'


def format_registers(output, prefix, registers):
    data = []
    for register in registers:
        data.append(('`{1} <{0}.{1}_>`_'.format(prefix, register.name), '0x{:02X}'.format(register.addressOffset)))
    if len(data):
        output.write(table(('Register', 'Offset'), data))

    for register in registers:
        output.write('.. _{}.{}:\n\n'.format(prefix, register.name))

        output.write(underline(register.description, section.subsubsection))

        output.write(rst_list_name.format('Name', register.name))
        output.write(rst_list_name.format('Size', register.size))
        output.write(rst_list_name.format('Offset', '0x{:02X}'.format(register.addressOffset)))
        output.write(rst_list_name.format('Reset', '0x{:{fill}{width}X}'.format(register.resetValue, fill='0', width=register.size // 4)))
        output.write(rst_list_name.format('Access', register.access))
        output.write('\n')

        # Table

        # Bit description
        for field in register.fields:
            if field.bitWidth == 1:
                output.write('- Bit {} ({}) - {}\n'.format(field.bitOffset, field.access, field.name))
            else:
                output.write('- Bits {}:{} ({}) - {}\n'.format(
                    field.bitOffset + field.bitWidth - 1, field.bitOffset, field.access, field.name))
            output.write('   {}\n\n'.format(field.description))

            if hasattr(field, 'enumeratedValues'):
                for enumeratedValue in field.enumeratedValues.enumeratedValues:
                    output.write('   - {} - {}\n'.format(enumeratedValue.value, enumeratedValue.name))
                    if hasattr(enumeratedValue, 'description'):
                        output.write('      {}\n'.format(enumeratedValue.description))
                output.write('\n')


def main():
    parser = argparse.ArgumentParser(description='SVD to ReST converter')
    parser.add_argument('--svd', metavar='FILE', type=str, help='System view description (SVD) file', required=True)
    parser.add_argument('--output', '-o',  metavar='FILE', type=str, help='ReST output file', required=True)
    parser.add_argument('--version', action='version', version=pysvd.__version__)
    args = parser.parse_args()

    node = ET.parse(args.svd).getroot()
    device = pysvd.element.Device(node)

    output = open(args.output, "w")

    # Device
    output.write(underline('Device', section.section))

    output.write(rst_list_name.format('Name', device.name))
    output.write(rst_list_name.format('Description', device.description))
    if hasattr(device, 'series'):
        output.write(rst_list_name.format('Series', device.series))
    output.write(rst_list_name.format('Version', device.version))
    if hasattr(device, 'vendor'):
        output.write(rst_list_name.format('Vendor', device.vendor))
    output.write('\n')

    output.write(rst_list_name.format('Address unit bits', device.addressUnitBits))
    output.write(rst_list_name.format('Data width', device.width))
    output.write('\n')

    # CPU
    output.write(underline('CPU', section.section))

    cpu = device.cpu
    output.write(rst_list_name.format('Name', str(cpu.name).replace('CM', 'Cortex-M').replace('CA', 'Cortex-A').replace('PLUS', '+')))
    output.write(rst_list_name.format('Revision', cpu.revision))
    output.write(rst_list_name.format('Endian', cpu.endian))
    output.write(rst_list_name.format('MPU', 'yes' if cpu.mpuPresent else 'no'))
    output.write(rst_list_name.format('FPU', 'yes' if cpu.fpuPresent else 'no'))
    if hasattr(cpu, 'fpuDP'):
        output.write(rst_list_name.format('FPU DP', 'yes' if cpu.fpuDP else 'no'))
    if hasattr(cpu, 'icachePresent'):
        output.write(rst_list_name.format('I-Cache', 'yes' if cpu.icachePresent else 'no'))
    if hasattr(cpu, 'dcachePresent'):
        output.write(rst_list_name.format('D-Cache', 'yes' if cpu.dcachePresent else 'no'))
    if hasattr(cpu, 'itcmPresent'):
        output.write(rst_list_name.format('ITCM', 'yes' if cpu.itcmPresent else 'no'))
    if hasattr(cpu, 'dtcmPresent'):
        output.write(rst_list_name.format('DTCM', 'yes' if cpu.dtcmPresent else 'no'))
    output.write(rst_list_name.format('VTOR', 'yes' if cpu.vtorPresent else 'no'))
    if hasattr(cpu, 'deviceNumInterrupts'):
        output.write(rst_list_name.format('Interrupts', cpu.deviceNumInterrupts))
    output.write(rst_list_name.format('Interrupt priorities', 2 ** cpu.nvicPrioBits))
    output.write(rst_list_name.format('Vendor SYSTICK', 'yes' if cpu.vendorSystickConfig else 'no'))
    output.write('\n')

    # Memory mapping
    output.write(underline('Memory mapping', section.section))

    data = []
    for peripheral in sorted(device.peripherals, key=lambda peripheral: peripheral.baseAddress):
        data.append(('{}_'.format(peripheral.name), '0x{:08X}'.format(peripheral.baseAddress)))
    output.write(table(('Peripheral', 'Address'), data))

    # Interrupt mapping
    output.write(underline('Interrupt mapping', section.section))

    data = []
    interrupts = []
    for peripheral in device.peripherals:
        interrupts += peripheral.interrupts

    for interrupt in sorted(interrupts, key=lambda interrupt: interrupt.value):
        data.append(('`{0}.{1} <{0}_>`_'.format(interrupt.parent.name, interrupt.name), str(interrupt.value)))
    output.write(table(('Peripheral', 'Interrupt'), data))

    # Peripheral
    output.write(underline('Peripheral', section.section))

    for peripheral in device.peripherals:
        output.write('.. _{}:\n\n'.format(peripheral.name))

        output.write(underline('{} ({})'.format(peripheral.description, peripheral.name), section.subsection))

        if hasattr(peripheral, 'version'):
            output.write(rst_list_name.format('Version', peripheral.version))
        output.write(rst_list_name.format('Address', '0x{:08X}'.format(peripheral.baseAddress)))
        for interrupt in peripheral.interrupts:
            output.write(rst_list_name.format('Interrupt {}'.format(interrupt.name), interrupt.value))
        output.write("\n")

        data = []
        for cluster in peripheral.clusters:
            data.append(('`{1} <{0}.{1}_>`_'.format(peripheral.name, cluster.name), '0x{:02X}'.format(cluster.addressOffset)))
        if len(data):
            output.write(table(('Cluster', 'Offset'), data))

        format_registers(output, peripheral.name, peripheral.registers)

        for cluster in peripheral.clusters:
            output.write('.. _{}.{}:\n\n'.format(peripheral.name, cluster.name))

            output.write(underline('{} ({})'.format(cluster.description, cluster.name), section.subsubsection))

            format_registers(output, '{}.{}'.format(peripheral.name, cluster.name), cluster.registers)

    output.write("Autogenerated ReST with pysvd {}\n".format(pysvd.__version__))
    output.close()


if __name__ == '__main__':
    main()



import unittest
import xml.etree.ElementTree as ET

import pysvd


class TestTreeComplete(unittest.TestCase):

    # def setup()
    # use setup to parse file and separate tests to check deriveFrom, derive  Attribute and dim functionallity

    @classmethod
    def setUpClass(cls):
        node = ET.parse("test/example.svd").getroot()
        cls.device = pysvd.element.Device(node)

    def test_device_attributes(self):
        device = self.device

        self.assertEqual(device.schemaVersion, '1.1')
        self.assertEqual(device.vendor, 'ARM Ltd.')
        self.assertEqual(device.vendorID, 'ARM')
        self.assertEqual(device.name, 'ARM_Example')
        self.assertEqual(device.series, 'ARMCM3')
        self.assertEqual(device.version, '1.2')
        self.assertEqual(device.description, 'ARM 32-bit Cortex-M3 Microcontroller based device, CPU clock up to 80MHz, etc.')
        self.assertEqual(device.addressUnitBits, 8)
        self.assertEqual(device.width, 32)
        self.assertEqual(device.size, 32)
        self.assertEqual(device.access, pysvd.type.access.read_write)
        self.assertEqual(device.resetValue, 0)
        self.assertEqual(device.resetMask, 0xffffffff)

    def test_cpu_attributes(self):
        cpu = self.device.cpu

        self.assertEqual(cpu.name, pysvd.type.cpuName.CM3)
        self.assertEqual(cpu.revision, 'r1p0')
        self.assertEqual(cpu.endian, pysvd.type.endian.little)
        self.assertTrue(cpu.mpuPresent)
        self.assertFalse(cpu.fpuPresent)
        self.assertEqual(cpu.nvicPrioBits, 3)
        self.assertFalse(cpu.vendorSystickConfig)

    def test_peripheral_equal_attributes(self):
        device = self.device

        self.assertEqual(len(device.peripherals), 3)
        for peripheral in device.peripherals:
            # general
            self.assertEqual(peripheral.version, "1.0")
            self.assertEqual(peripheral.size, 32)
            self.assertEqual(peripheral.access, pysvd.type.access.read_write)
            self.assertEqual(peripheral.groupName, "TIMER")

            # inherited
            self.assertEqual(peripheral.resetValue, 0x00000000)
            self.assertEqual(peripheral.resetMask, 0xffffffff)

            # addressBlock
            self.assertEqual(len(peripheral.addressBlocks), 1)
            self.assertEqual(peripheral.addressBlocks[0].offset, 0)
            self.assertEqual(peripheral.addressBlocks[0].size, 0x100)
            self.assertEqual(peripheral.addressBlocks[0].usage, pysvd.type.addressBlockUsage.registers)

    def test_peripheral_derivedFrom_attributes(self):
        device = self.device

        peripheral_index = 0
        self.assertEqual(len(device.peripherals), 3)
        for peripheral in device.peripherals:
            if peripheral_index == 0:
                # general
                self.assertEqual(peripheral.name, "TIMER0")
                self.assertEqual(peripheral.description, "32 Timer / Counter, counting up or down from different sources")
                self.assertEqual(peripheral.baseAddress, 0x40010000)
                self.assertIsNone(peripheral.derivedFrom)

                # interrupt
                self.assertEqual(len(peripheral.interrupts), 1)
                self.assertEqual(peripheral.interrupts[0].name, "TIMER0")
                self.assertEqual(peripheral.interrupts[0].description, "Timer 0 interrupt")
                self.assertEqual(peripheral.interrupts[0].value, 0)

            elif peripheral_index == 1:
                # general
                self.assertEqual(peripheral.name, "TIMER1")
                self.assertEqual(peripheral.description, "32 Timer / Counter, counting up or down from different sources")
                self.assertEqual(peripheral.baseAddress, 0x40010100)
                self.assertEqual(peripheral.derivedFrom, device.peripherals[0])

                # interrupt
                self.assertEqual(len(peripheral.interrupts), 1)
                self.assertEqual(peripheral.interrupts[0].name, "TIMER1")
                self.assertEqual(peripheral.interrupts[0].description, "Timer 1 interrupt")
                self.assertEqual(peripheral.interrupts[0].value, 4)

            elif peripheral_index == 2:
                # general
                self.assertEqual(peripheral.name, "TIMER2")
                self.assertEqual(peripheral.description, "32 Timer / Counter, counting up or down from different sources")
                self.assertEqual(peripheral.baseAddress, 0x40010200)
                self.assertEqual(peripheral.derivedFrom, device.peripherals[0])

                # interrupt
                self.assertEqual(len(peripheral.interrupts), 1)
                self.assertEqual(peripheral.interrupts[0].name, "TIMER2")
                self.assertEqual(peripheral.interrupts[0].description, "Timer 2 interrupt")
                self.assertEqual(peripheral.interrupts[0].value, 6)

            # registers
            register_index = 0
            self.assertEqual(len(peripheral.registers), 8)
            for register in peripheral.registers:
                if register_index == 0:
                    self.assertEqual(register.name, "CR")
                    self.assertEqual(register.description, "Control Register")
                    self.assertEqual(register.addressOffset, 0)
                    self.assertEqual(register.size, 32)
                    self.assertEqual(register.access, pysvd.type.access.read_write)
                    self.assertEqual(register.resetValue, 0x00000000)
                    self.assertEqual(register.resetMask, 0x1337f7f)

                    # fields
                    field_index = 0
                    self.assertEqual(len(register.fields), 12)
                    for field in register.fields:
                        if field_index == 0:
                            self.assertEqual(field.name, "EN")
                            self.assertEqual(field.description, "Enable")
                            self.assertEqual(field.bitOffset, 0)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "Disable")
                                    self.assertEqual(enumeratedValue.description, "Timer is disabled and does not operate")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Enable")
                                    self.assertEqual(enumeratedValue.description, "Timer is enabled and can operate")
                                    self.assertEqual(enumeratedValue.value, 1)
                                enumeratedValue_index += 1

                        elif field_index == 1:
                            self.assertEqual(field.name, "RST")
                            self.assertEqual(field.description, "Reset Timer")
                            self.assertEqual(field.bitOffset, 1)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.write_only)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "No_Action")
                                    self.assertEqual(enumeratedValue.description, "Write as ZERO if necessary")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Reset_Timer")
                                    self.assertEqual(enumeratedValue.description, "Reset the Timer")
                                    self.assertEqual(enumeratedValue.value, 1)

                                enumeratedValue_index += 1

                        elif field_index == 2:
                            self.assertEqual(field.name, "CNT")
                            self.assertEqual(field.description, "Counting direction")
                            self.assertEqual(field.bitOffset, 2)
                            self.assertEqual(field.bitWidth, 2)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 3)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "Count_UP")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer Counts UO and wraps, if no STOP condition is set")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Count_DOWN")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer Counts DOWN and wraps, if no STOP condition is set")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 2:
                                    self.assertEqual(enumeratedValue.name, "Toggle")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer Counts up to MAX, then DOWN to ZERO, if no STOP condition is set")
                                    self.assertEqual(enumeratedValue.value, 2)

                                enumeratedValue_index += 1

                        elif field_index == 3:
                            self.assertEqual(field.name, "MODE")
                            self.assertEqual(field.description, "Operation Mode")
                            self.assertEqual(field.bitOffset, 4)
                            self.assertEqual(field.bitWidth, 3)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 5)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "Continous")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer runs continously")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Single_ZERO_MAX")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer counts to 0x00 or 0xFFFFFFFF (depending on CNT) and stops")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 2:
                                    self.assertEqual(enumeratedValue.name, "Single_MATCH")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer counts to the Value of MATCH Register and stops")
                                    self.assertEqual(enumeratedValue.value, 2)

                                elif enumeratedValue_index == 3:
                                    self.assertEqual(enumeratedValue.name, "Reload_ZERO_MAX")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer counts to 0x00 or 0xFFFFFFFF (depending on CNT), loads the RELOAD Value and continues")
                                    self.assertEqual(enumeratedValue.value, 3)

                                elif enumeratedValue_index == 4:
                                    self.assertEqual(enumeratedValue.name, "Reload_MATCH")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer counts to the Value of MATCH Register, loads the RELOAD Value and continues")
                                    self.assertEqual(enumeratedValue.value, 4)

                                enumeratedValue_index += 1

                        elif field_index == 4:
                            self.assertEqual(field.name, "PSC")
                            self.assertEqual(field.description, "Use Prescaler")
                            self.assertEqual(field.bitOffset, 7)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "Disabled")
                                    self.assertEqual(enumeratedValue.description, "Prescaler is not used")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Enabled")
                                    self.assertEqual(enumeratedValue.description, "Prescaler is used as divider")
                                    self.assertEqual(enumeratedValue.value, 1)

                                enumeratedValue_index += 1

                        elif field_index == 5:
                            self.assertEqual(field.name, "CNTSRC")
                            self.assertEqual(field.description, "Timer / Counter Source Divider")
                            self.assertEqual(field.bitOffset, 8)
                            self.assertEqual(field.bitWidth, 4)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 9)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "CAP_SRC")
                                    self.assertEqual(enumeratedValue.description, "Capture Source is used directly")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "CAP_SRC_div2")
                                    self.assertEqual(enumeratedValue.description, "Capture Source is divided by 2")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 2:
                                    self.assertEqual(enumeratedValue.name, "CAP_SRC_div4")
                                    self.assertEqual(enumeratedValue.description, "Capture Source is divided by 4")
                                    self.assertEqual(enumeratedValue.value, 2)

                                elif enumeratedValue_index == 3:
                                    self.assertEqual(enumeratedValue.name, "CAP_SRC_div8")
                                    self.assertEqual(enumeratedValue.description, "Capture Source is divided by 8")
                                    self.assertEqual(enumeratedValue.value, 3)

                                elif enumeratedValue_index == 4:
                                    self.assertEqual(enumeratedValue.name, "CAP_SRC_div16")
                                    self.assertEqual(enumeratedValue.description, "Capture Source is divided by 16")
                                    self.assertEqual(enumeratedValue.value, 4)

                                elif enumeratedValue_index == 5:
                                    self.assertEqual(enumeratedValue.name, "CAP_SRC_div32")
                                    self.assertEqual(enumeratedValue.description, "Capture Source is divided by 32")
                                    self.assertEqual(enumeratedValue.value, 5)

                                elif enumeratedValue_index == 6:
                                    self.assertEqual(enumeratedValue.name, "CAP_SRC_div64")
                                    self.assertEqual(enumeratedValue.description, "Capture Source is divided by 64")
                                    self.assertEqual(enumeratedValue.value, 6)

                                elif enumeratedValue_index == 7:
                                    self.assertEqual(enumeratedValue.name, "CAP_SRC_div128")
                                    self.assertEqual(enumeratedValue.description, "Capture Source is divided by 128")
                                    self.assertEqual(enumeratedValue.value, 7)

                                elif enumeratedValue_index == 8:
                                    self.assertEqual(enumeratedValue.name, "CAP_SRC_div256")
                                    self.assertEqual(enumeratedValue.description, "Capture Source is divided by 256")
                                    self.assertEqual(enumeratedValue.value, 8)

                                enumeratedValue_index += 1

                        elif field_index == 6:
                            self.assertEqual(field.name, "CAPSRC")
                            self.assertEqual(field.description, "Timer / Counter Capture Source")
                            self.assertEqual(field.bitOffset, 12)
                            self.assertEqual(field.bitWidth, 4)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 16)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "CClk")
                                    self.assertEqual(enumeratedValue.description, "Core Clock")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "GPIOA_0")
                                    self.assertEqual(enumeratedValue.description, "GPIO A, PIN 0")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 15:
                                    self.assertEqual(enumeratedValue.name, "GPIOC_6")
                                    self.assertEqual(enumeratedValue.description, "GPIO C, PIN 2")
                                    self.assertEqual(enumeratedValue.value, 15)

                                enumeratedValue_index += 1

                        elif field_index == 7:
                            self.assertEqual(field.name, "CAPEDGE")
                            self.assertEqual(
                                field.description,
                                "Capture Edge, select which Edge should result in a counter increment or decrement")
                            self.assertEqual(field.bitOffset, 16)
                            self.assertEqual(field.bitWidth, 2)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 3)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "RISING")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Only rising edges result in a counter increment or decrement")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "FALLING")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Only falling edges result in a counter increment or decrement")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 2:
                                    self.assertEqual(enumeratedValue.name, "BOTH")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Rising and falling edges result in a counter increment or decrement")
                                    self.assertEqual(enumeratedValue.value, 2)

                                enumeratedValue_index += 1

                        elif field_index == 8:
                            self.assertEqual(field.name, "TRGEXT")
                            self.assertEqual(field.description, "Triggers an other Peripheral")
                            self.assertEqual(field.bitOffset, 20)
                            self.assertEqual(field.bitWidth, 2)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 4)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "NONE")
                                    self.assertEqual(enumeratedValue.description, "No Trigger is emitted")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "DMA1")
                                    self.assertEqual(enumeratedValue.description, "DMA Controller 1 is triggered, dependant on MODE")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 2:
                                    self.assertEqual(enumeratedValue.name, "DMA2")
                                    self.assertEqual(enumeratedValue.description, "DMA Controller 2 is triggered, dependant on MODE")
                                    self.assertEqual(enumeratedValue.value, 2)

                                elif enumeratedValue_index == 3:
                                    self.assertEqual(enumeratedValue.name, "UART")
                                    self.assertEqual(enumeratedValue.description, "UART is triggered, dependant on MODE")
                                    self.assertEqual(enumeratedValue.value, 3)

                                enumeratedValue_index += 1

                        elif field_index == 9:
                            self.assertEqual(field.name, "RELOAD")
                            self.assertEqual(field.description, "Select RELOAD Register n to reload Timer on condition")
                            self.assertEqual(field.bitOffset, 24)
                            self.assertEqual(field.bitWidth, 2)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 4)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "RELOAD0")
                                    self.assertEqual(enumeratedValue.description, "Selects Reload Register number 0")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "RELOAD1")
                                    self.assertEqual(enumeratedValue.description, "Selects Reload Register number 1")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 2:
                                    self.assertEqual(enumeratedValue.name, "RELOAD2")
                                    self.assertEqual(enumeratedValue.description, "Selects Reload Register number 2")
                                    self.assertEqual(enumeratedValue.value, 2)

                                elif enumeratedValue_index == 3:
                                    self.assertEqual(enumeratedValue.name, "RELOAD3")
                                    self.assertEqual(enumeratedValue.description, "Selects Reload Register number 3")
                                    self.assertEqual(enumeratedValue.value, 3)

                                enumeratedValue_index += 1

                        elif field_index == 10:
                            self.assertEqual(field.name, "IDR")
                            self.assertEqual(
                                field.description,
                                "Selects, if Reload Register number is incremented, decremented or not modified")
                            self.assertEqual(field.bitOffset, 26)
                            self.assertEqual(field.bitWidth, 2)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 3)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "KEEP")
                                    self.assertEqual(enumeratedValue.description, "Reload Register number does not change automatically")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "INCREMENT")
                                    self.assertEqual(enumeratedValue.description, "Reload Register number is incremented on each match")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 2:
                                    self.assertEqual(enumeratedValue.name, "DECREMENT")
                                    self.assertEqual(enumeratedValue.description, "Reload Register number is decremented on each match")
                                    self.assertEqual(enumeratedValue.value, 2)

                                enumeratedValue_index += 1

                        elif field_index == 11:
                            self.assertEqual(field.name, "S")
                            self.assertEqual(field.description, "Starts and Stops the Timer / Counter")
                            self.assertEqual(field.bitOffset, 31)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "STOP")
                                    self.assertEqual(enumeratedValue.description, "Timer / Counter is stopped")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "START")
                                    self.assertEqual(enumeratedValue.description, "Timer / Counter is started")
                                    self.assertEqual(enumeratedValue.value, 1)

                                enumeratedValue_index += 1

                        field_index += 1

                elif register_index == 1:
                    self.assertEqual(register.name, "SR")
                    self.assertEqual(register.description, "Status Register")
                    self.assertEqual(register.addressOffset, 0x04)
                    self.assertEqual(register.size, 16)
                    self.assertEqual(register.access, pysvd.type.access.read_write)
                    self.assertEqual(register.resetValue, 0x0000)
                    self.assertEqual(register.resetMask, 0xd701)

                    # fields
                    field_index = 0
                    self.assertEqual(len(register.fields), 6)
                    for field in register.fields:
                        if field_index == 0:
                            self.assertEqual(field.name, "RUN")
                            self.assertEqual(field.description, "Shows if Timer is running or not")
                            self.assertEqual(field.bitOffset, 0)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.read_only)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "Stopped")
                                    self.assertEqual(enumeratedValue.description, "Timer is not running")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Running")
                                    self.assertEqual(enumeratedValue.description, "Timer is running")
                                    self.assertEqual(enumeratedValue.value, 1)

                                enumeratedValue_index += 1

                        elif field_index == 1:
                            self.assertEqual(field.name, "MATCH")
                            self.assertEqual(field.description, "Shows if the MATCH was hit")
                            self.assertEqual(field.bitOffset, 8)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "No_Match")
                                    self.assertEqual(enumeratedValue.description, "The MATCH condition was not hit")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Match_Hit")
                                    self.assertEqual(enumeratedValue.description, "The MATCH condition was hit")
                                    self.assertEqual(enumeratedValue.value, 1)

                                enumeratedValue_index += 1

                        elif field_index == 2:
                            self.assertEqual(field.name, "UN")
                            self.assertEqual(field.description, "Shows if an underflow occured. This flag is sticky")
                            self.assertEqual(field.bitOffset, 9)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "No_Underflow")
                                    self.assertEqual(enumeratedValue.description, "No underflow occured since last clear")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Underflow")
                                    self.assertEqual(enumeratedValue.description, "A minimum of one underflow occured since last clear")
                                    self.assertEqual(enumeratedValue.value, 1)

                                enumeratedValue_index += 1

                        elif field_index == 3:
                            self.assertEqual(field.name, "OV")
                            self.assertEqual(field.description, "Shows if an overflow occured. This flag is sticky")
                            self.assertEqual(field.bitOffset, 10)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "No_Overflow")
                                    self.assertEqual(enumeratedValue.description, "No overflow occured since last clear")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Overflow_occured")
                                    self.assertEqual(enumeratedValue.description, "A minimum of one overflow occured since last clear")
                                    self.assertEqual(enumeratedValue.value, 1)

                                enumeratedValue_index += 1

                        elif field_index == 4:
                            self.assertEqual(field.name, "RST")
                            self.assertEqual(field.description, "Shows if Timer is in RESET state")
                            self.assertEqual(field.bitOffset, 12)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.read_only)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "Ready")
                                    self.assertEqual(enumeratedValue.description, "Timer is not in RESET state and can operate")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "In_Reset")
                                    self.assertEqual(enumeratedValue.description, "Timer is in RESET state and can not operate")
                                    self.assertEqual(enumeratedValue.value, 1)

                                enumeratedValue_index += 1

                        elif field_index == 5:
                            self.assertEqual(field.name, "RELOAD")
                            self.assertEqual(field.description, "Shows the currently active RELOAD Register")
                            self.assertEqual(field.bitOffset, 14)
                            self.assertEqual(field.bitWidth, 2)
                            self.assertEqual(field.access, pysvd.type.access.read_only)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 4)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "RELOAD0")
                                    self.assertEqual(enumeratedValue.description, "Reload Register number 0 is active")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "RELOAD1")
                                    self.assertEqual(enumeratedValue.description, "Reload Register number 1 is active")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 2:
                                    self.assertEqual(enumeratedValue.name, "RELOAD2")
                                    self.assertEqual(enumeratedValue.description, "Reload Register number 2 is active")
                                    self.assertEqual(enumeratedValue.value, 2)

                                elif enumeratedValue_index == 3:
                                    self.assertEqual(enumeratedValue.name, "RELOAD3")
                                    self.assertEqual(enumeratedValue.description, "Reload Register number 3 is active")
                                    self.assertEqual(enumeratedValue.value, 3)

                                enumeratedValue_index += 1

                        field_index += 1

                elif register_index == 2:
                    self.assertEqual(register.name, "INT")
                    self.assertEqual(register.description, "Interrupt Register")
                    self.assertEqual(register.addressOffset, 0x10)
                    self.assertEqual(register.size, 16)
                    self.assertEqual(register.access, pysvd.type.access.read_write)
                    self.assertEqual(register.resetValue, 0x0000)
                    self.assertEqual(register.resetMask, 0x0771)

                    # fields
                    field_index = 0
                    self.assertEqual(len(register.fields), 2)
                    for field in register.fields:
                        if field_index == 0:
                            self.assertEqual(field.name, "EN")
                            self.assertEqual(field.description, "Interrupt Enable")
                            self.assertEqual(field.bitOffset, 0)
                            self.assertEqual(field.bitWidth, 1)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 2)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "Disabled")
                                    self.assertEqual(enumeratedValue.description, "Timer does not generate Interrupts")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Enable")
                                    self.assertEqual(enumeratedValue.description, "Timer triggers the TIMERn Interrupt")
                                    self.assertEqual(enumeratedValue.value, 1)

                                enumeratedValue_index += 1

                        elif field_index == 1:
                            self.assertEqual(field.name, "MODE")
                            self.assertEqual(
                                field.description,
                                "Interrupt Mode, selects on which condition the Timer should generate an Interrupt")
                            self.assertEqual(field.bitOffset, 4)
                            self.assertEqual(field.bitWidth, 3)
                            self.assertEqual(field.access, pysvd.type.access.read_write)

                            # enumeratedValues
                            enumeratedValue_index = 0
                            self.assertEqual(len(field.enumeratedValues.enumeratedValues), 3)
                            for enumeratedValue in field.enumeratedValues.enumeratedValues:
                                if enumeratedValue_index == 0:
                                    self.assertEqual(enumeratedValue.name, "Match")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer generates an Interrupt when the MATCH condition is hit")
                                    self.assertEqual(enumeratedValue.value, 0)

                                elif enumeratedValue_index == 1:
                                    self.assertEqual(enumeratedValue.name, "Underflow")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer generates an Interrupt when it underflows")
                                    self.assertEqual(enumeratedValue.value, 1)

                                elif enumeratedValue_index == 2:
                                    self.assertEqual(enumeratedValue.name, "Overflow")
                                    self.assertEqual(
                                        enumeratedValue.description,
                                        "Timer generates an Interrupt when it overflows")
                                    self.assertEqual(enumeratedValue.value, 2)

                                enumeratedValue_index += 1

                        field_index += 1

                elif register_index == 3:
                    self.assertEqual(register.name, "COUNT")
                    self.assertEqual(register.description, "The Counter Register reflects the actual Value of the Timer/Counter")
                    self.assertEqual(register.addressOffset, 0x20)
                    self.assertEqual(register.size, 32)
                    self.assertEqual(register.access, pysvd.type.access.read_write)
                    self.assertEqual(register.resetValue, 0x00000000)
                    self.assertEqual(register.resetMask, 0xffffffff)

                elif register_index == 4:
                    self.assertEqual(register.name, "MATCH")
                    self.assertEqual(register.description, "The Match Register stores the compare Value for the MATCH condition")
                    self.assertEqual(register.addressOffset, 0x24)
                    self.assertEqual(register.size, 32)
                    self.assertEqual(register.access, pysvd.type.access.read_write)
                    self.assertEqual(register.resetValue, 0x00000000)
                    self.assertEqual(register.resetMask, 0xffffffff)

                elif register_index == 5:
                    self.assertEqual(register.name, "PRESCALE_RD")
                    self.assertEqual(
                        register.description,
                        "The Prescale Register stores the Value for the prescaler. The cont event gets divided by this value")
                    self.assertEqual(register.addressOffset, 0x28)
                    self.assertEqual(register.size, 32)
                    self.assertEqual(register.access, pysvd.type.access.read_only)
                    self.assertEqual(register.resetValue, 0x00000000)
                    self.assertEqual(register.resetMask, 0xffffffff)

                elif register_index == 6:
                    self.assertEqual(register.name, "PRESCALE_WR")
                    self.assertEqual(
                        register.description,
                        "The Prescale Register stores the Value for the prescaler. The cont event gets divided by this value")
                    self.assertEqual(register.addressOffset, 0x28)
                    self.assertEqual(register.size, 32)
                    self.assertEqual(register.access, pysvd.type.access.write_only)
                    self.assertEqual(register.resetValue, 0x00000000)
                    self.assertEqual(register.resetMask, 0xffffffff)

                elif register_index == 7:
                    self.assertEqual(register.name, "RELOAD[4]")
                    self.assertEqual(
                        register.description,
                        "The Reload Register stores the Value the COUNT Register gets reloaded on a when a condition was met.")
                    self.assertEqual(register.addressOffset, 0x50)
                    self.assertEqual(register.size, 32)
                    self.assertEqual(register.access, pysvd.type.access.read_write)
                    self.assertEqual(register.resetValue, 0x00000000)
                    self.assertEqual(register.resetMask, 0xffffffff)

                register_index += 1

            peripheral_index += 1



import re
import pysvd

# Note construtors: First class specific code is executed than parent constructor
# called, so that on last constructor of Base parser() can be called also hierachical
# on every parent object automatically.


class Base(object):
    """Base class for all SVD elements"""

    def __init__(self, node):
        self.node = node
        self.parent = getattr(self, 'parent', None)
        self.derivedFrom = None

        self.parse(self.node)

    def parse(self, node):
        """Overwrite in derived classes to parse nodes"""
        pass

    def add_attribute(self, node, name, parser_type, mandatory=False, default=None):
        """Parse node element as given type and add it to self if not None"""
        value = parser_type(pysvd.node.Element(node, name, mandatory), default)
        if value is not None:
            self.__dict__[name] = value

    def add_enum_attribute(self, node, name, enum, mandatory=False, default=None):
        """Parse node element as given enum and add it to self if not None"""
        value = pysvd.parser.Enum(enum, pysvd.node.Element(node, name, mandatory), default)
        if value is not None:
            self.__dict__[name] = value

    def find(self, name):
        """Find child by name. Has to be overwritten by each derived class with child elements."""
        assert not hasattr(super(), 'find')
        return None

    @classmethod
    def add_elements(cls, parent, elements, node, name):
        """Parse node elements and add them to elements list"""

        for subnode in node.findall(name):
            elements.append(cls(parent, subnode))


class Parent(Base):
    """Base class for parents"""

    def __init__(self, parent, node):
        self.parent = parent

        super().__init__(node)


class Group(Parent):
    """Base class for elements with registerPropertiesGroup"""

    attributes = ['size', 'access', 'protection', 'resetValue', 'resetMask']

    def __init__(self, parent_, node):
        super().__init__(parent_, node)

    def __getattr__(self, attr):
        if attr in self.attributes:
            parent = self.parent
            while parent is not None:
                try:
                    return parent.__getattribute__(attr)
                except AttributeError:
                    parent = parent.parent

        raise AttributeError("'{}' object has no attribute '{}'".format(self.__class__.__name__, attr))


class Derive(Group):
    """Base for deriveable classes"""

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        # If derived, search class, call parse attributes of derived object and call base ctor
        derivedFrom = pysvd.node.Attribute(node, 'derivedFrom')
        if derivedFrom is not None:
            parts = derivedFrom.split('.')
            count = len(parts) - 1
            object = self.parent
            while count:
                object = object.parent
                count -= 1

            if object is None:
                name = pysvd.parser.Text(pysvd.node.Element(node, 'name'))
                raise KeyError("Can not find root element from path '{}' to derive '{}'".format(derivedFrom, name))

            for name in parts:
                res = object.find(name)
                if res is None:
                    raise KeyError("Can not find path element '{}' from path '{}' in object '{}'".format(name, derivedFrom, object.name))
                object = res

            self.parse(object.node)
            self.derivedFrom = object


class Dim(Derive):

    def __init__(self, parent, node):
        super().__init__(parent, node)

    def parse(self, node):
        super().parse(node)

        self.name = pysvd.parser.Text(pysvd.node.Element(node, 'name', True))
        displayName = pysvd.parser.Text(pysvd.node.Element(node, 'displayName'))
        if displayName is not None:
            self.displayName = displayName
        description = pysvd.parser.Text(pysvd.node.Element(node, 'description'))
        if description is not None:
            self.description = description
        self.dimName = pysvd.parser.Text(pysvd.node.Element(node, 'dimName'), self.name)

    # Replace %s with name if not None
    def set_index(self, value):
        value = str(value)
        self.name = self.name.replace('%s', value)
        if hasattr(self, 'displayName') and self.displayName is not None:
            self.displayName = self.displayName.replace('%s', value)
        if hasattr(self, 'description') and self.description is not None:
            self.description = self.description.replace('%s', value)

        if self.dimName is not None:
            self.dimName = self.dimName.replace('%s', value)

    def set_offset(self, value):
        pass

    @classmethod
    def add_elements(cls, parent, elements, node, name):
        """Parse node elements with respect to dim entries and return a list with constucted elements"""

        for subnode in node.findall(name):
            dim = pysvd.parser.Integer(pysvd.node.Element(subnode, 'dim'))
            if dim is not None:
                dimIncrement = pysvd.parser.Integer(pysvd.node.Element(subnode, 'dimIncrement', True))
                dimIndex = pysvd.parser.Text(pysvd.node.Element(subnode, 'dimIndex'))

                # if dimIndex is not present, dimName and name has to be examined for '[%s]' string presence,
                # to distinguish between array and index
                if dimIndex is None:
                    dimName = pysvd.parser.Text(pysvd.node.Element(subnode, 'dimName'))
                    name = pysvd.parser.Text(pysvd.node.Element(subnode, 'name'))
                    if dimName is not None and '[%s]' not in dimName or name is not None and '[%s]' not in name:
                        dimIndices = range(dim)
                    else:
                        dimIndices = [dim]
                else:
                    if ',' in dimIndex:
                        dimIndices = dimIndex.split(',')
                    elif '-' in dimIndex:
                        match = re.search(r'([0-9]+)\-([0-9]+)', dimIndex)
                        dimIndices = list(range(int(match.group(1)), int(match.group(2)) + 1))
                    else:
                        raise ValueError("Unexpected value in 'dimIndex': {}".format(dimIndex))

                    if len(dimIndices) != dim:
                        raise AttributeError("'dim' size does not match elements in 'dimIndex' ({} != {})".format(dim, len(dimIndex)))

                offset = 0
                for index in dimIndices:
                    object = cls(parent, subnode)
                    object.set_index(index)
                    object.set_offset(offset)
                    elements.append(object)
                    offset += dimIncrement
            else:
                elements.append(cls(parent, subnode))



import setuptools
import unittest

exec(open('pysvd/version.py').read())

with open("README.md", "r") as file:
    long_description = file.read()

setuptools.setup(
    name="pysvd",
    version=__version__,
    description="System View Description parser",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="SVD CMSIS ARM",
    license="MIT",
    author="Benjamin Füldner",
    author_email="benjamin@fueldner.net",
    url="https://code.fueldner.net/opensource/pysvd",
    project_urls={
        'Bug Reports': 'https://github.com/bfueldner/pysvd/issues',
        'Source': 'https://github.com/bfueldner/pysvd',
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: System :: Hardware",
        "Topic :: System :: Hardware :: Hardware Drivers",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Embedded Systems",
    ],
    python_requires='>=3.5',
    entry_points={
        'console_scripts': [
            'svd2rst = scripts.svd2rst:main',
            'svd2register = scripts.svd2register:main',
            'svd_duplicates = scripts.svd_duplicates:main',
        ],
    },
    setup_requires=["pytest-runner"],
    tests_require=["pytest"]
)



import unittest
import xml.etree.ElementTree as ET

import pysvd


class TestSpecialCluster(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        node = ET.parse("test/specialCluster.xml").getroot()
        cls.peripheral = pysvd.element.Peripheral(None, node)

    def test_general(self):
        peripheral = self.peripheral

        self.assertEqual(peripheral.name, "RTC")
        self.assertEqual(peripheral.version, "1.0.1")
        self.assertEqual(peripheral.description, "Real-Time Counter")
        self.assertEqual(peripheral.groupName, "RTC")
        self.assertEqual(peripheral.prependToName, "RTC_")
        self.assertEqual(peripheral.baseAddress, 0x40001400)

    def test_address_block(self):
        self.assertEqual(len(self.peripheral.addressBlocks), 1)
        addressBlock = self.peripheral.addressBlocks[0]

        self.assertEqual(addressBlock.offset, 0)
        self.assertEqual(addressBlock.size, 0x40)
        self.assertEqual(addressBlock.usage, pysvd.type.addressBlockUsage.registers)

    def test_interrupt(self):
        self.assertEqual(len(self.peripheral.interrupts), 1)
        interrupt = self.peripheral.interrupts[0]

        self.assertEqual(interrupt.name, "RTC_INTREQ")
        self.assertEqual(interrupt.value, 3)

    def test_register(self):
        peripheral = self.peripheral

        self.assertEqual(len(peripheral.registers), 0)

    def test_cluster(self):
        peripheral = self.peripheral

        cluster_index = 0
        self.assertEqual(len(peripheral.clusters), 3)
        for cluster in peripheral.clusters:
            if cluster_index == 0:
                self.assertEqual(cluster.name, "MODE0")
                self.assertEqual(cluster.description, "32-bit Counter with Single 32-bit Compare")
                self.assertEqual(cluster.headerStructName, "RtcMode0")
                self.assertEqual(cluster.addressOffset, 0)

            elif cluster_index == 1:
                self.assertEqual(cluster.name, "MODE1")
                self.assertEqual(cluster.description, "16-bit Counter with Two 16-bit Compares")
                self.assertEqual(cluster.alternateCluster, "MODE0")
                self.assertEqual(cluster.headerStructName, "RtcMode1")
                self.assertEqual(cluster.addressOffset, 0)

            elif cluster_index == 2:
                self.assertEqual(cluster.name, "MODE2")
                self.assertEqual(cluster.description, "Clock/Calendar with Alarm")
                self.assertEqual(cluster.alternateCluster, "MODE0")
                self.assertEqual(cluster.headerStructName, "RtcMode2")
                self.assertEqual(cluster.addressOffset, 0)

            cluster_index += 1

    def test_cluster0_register(self):
        peripheral = self.peripheral

        self.assertEqual(len(peripheral.clusters), 3)
        cluster = peripheral.clusters[0]

        register_index = 0
        self.assertEqual(len(cluster.registers), 5)
        for register in cluster.registers:
            if register_index == 0:
                self.assertEqual(register.name, "CTRL")
                self.assertEqual(register.description, "MODE0 Control")
                self.assertEqual(register.addressOffset, 0x00)
                self.assertEqual(register.size, 16)

            elif register_index == 1:
                self.assertEqual(register.name, "READREQ")
                self.assertEqual(register.description, "Read Request")
                self.assertEqual(register.addressOffset, 0x02)
                self.assertEqual(register.size, 16)
                self.assertEqual(register.resetValue, 0x0010)

            elif register_index == 2:
                self.assertEqual(register.name, "STATUS")
                self.assertEqual(register.description, "Status")
                self.assertEqual(register.addressOffset, 0x0A)
                self.assertEqual(register.size, 8)

            elif register_index == 3:
                self.assertEqual(register.name, "COUNT")
                self.assertEqual(register.description, "MODE0 Counter Value")
                self.assertEqual(register.addressOffset, 0x10)
                self.assertEqual(register.size, 32)

            elif register_index == 4:
                self.assertEqual(register.name, "COMP0")
                self.assertEqual(register.description, "MODE0 Compare n Value")
                self.assertEqual(register.addressOffset, 0x18)
                self.assertEqual(register.size, 32)

            register_index += 1

    def test_cluster1_register(self):
        peripheral = self.peripheral

        self.assertEqual(len(peripheral.clusters), 3)
        cluster = peripheral.clusters[1]

        register_index = 0
        self.assertEqual(len(cluster.registers), 7)
        for register in cluster.registers:
            if register_index == 0:
                self.assertEqual(register.name, "CTRL")
                self.assertEqual(register.description, "MODE1 Control")
                self.assertEqual(register.addressOffset, 0x00)
                self.assertEqual(register.size, 16)

            elif register_index == 1:
                self.assertEqual(register.name, "READREQ")
                self.assertEqual(register.description, "Read Request")
                self.assertEqual(register.addressOffset, 0x02)
                self.assertEqual(register.size, 16)
                self.assertEqual(register.resetValue, 0x0010)

            elif register_index == 2:
                self.assertEqual(register.name, "STATUS")
                self.assertEqual(register.description, "Status")
                self.assertEqual(register.addressOffset, 0x0A)
                self.assertEqual(register.size, 8)

            elif register_index == 3:
                self.assertEqual(register.name, "COUNT")
                self.assertEqual(register.description, "MODE1 Counter Value")
                self.assertEqual(register.addressOffset, 0x10)
                self.assertEqual(register.size, 16)

            elif register_index == 4:
                self.assertEqual(register.name, "PER")
                self.assertEqual(register.description, "MODE1 Counter Period")
                self.assertEqual(register.addressOffset, 0x14)
                self.assertEqual(register.size, 16)

            elif register_index == 5:
                self.assertEqual(register.name, "COMP0")
                self.assertEqual(register.description, "MODE1 Compare n Value")
                self.assertEqual(register.addressOffset, 0x18)
                self.assertEqual(register.size, 16)

            elif register_index == 6:
                self.assertEqual(register.name, "COMP1")
                self.assertEqual(register.description, "MODE1 Compare n Value")
                self.assertEqual(register.addressOffset, 0x1A)
                self.assertEqual(register.size, 16)

            register_index += 1

    def test_cluster2_register(self):
        peripheral = self.peripheral

        self.assertEqual(len(peripheral.clusters), 3)
        cluster = peripheral.clusters[2]

        register_index = 0
        self.assertEqual(len(cluster.registers), 6)
        for register in cluster.registers:
            if register_index == 0:
                self.assertEqual(register.name, "CTRL")
                self.assertEqual(register.description, "MODE2 Control")
                self.assertEqual(register.addressOffset, 0x00)
                self.assertEqual(register.size, 16)

            elif register_index == 1:
                self.assertEqual(register.name, "READREQ")
                self.assertEqual(register.description, "Read Request")
                self.assertEqual(register.addressOffset, 0x02)
                self.assertEqual(register.size, 16)
                self.assertEqual(register.resetValue, 0x0010)

            elif register_index == 2:
                self.assertEqual(register.name, "STATUS")
                self.assertEqual(register.description, "Status")
                self.assertEqual(register.addressOffset, 0x0A)
                self.assertEqual(register.size, 8)

            elif register_index == 3:
                self.assertEqual(register.name, "CLOCK")
                self.assertEqual(register.description, "MODE2 Clock Value")
                self.assertEqual(register.addressOffset, 0x10)
                self.assertEqual(register.size, 32)

            elif register_index == 4:
                self.assertEqual(register.name, "ALARM0")
                self.assertEqual(register.description, "MODE2 Alarm n Value")
                self.assertEqual(register.addressOffset, 0x18)
                self.assertEqual(register.size, 32)

            elif register_index == 5:
                self.assertEqual(register.name, "MASK0")
                self.assertEqual(register.description, "MODE2 Alarm n Mask")
                self.assertEqual(register.addressOffset, 0x1C)
                self.assertEqual(register.size, 8)

            register_index += 1


class TestSpecialDerivePeripheral(unittest.TestCase):

    def test_ctor(self):
        pass

    def test_attributes(self):
        pass


class TestSpecialDeriveRegister(unittest.TestCase):

    def test_ctor(self):
        pass

    def test_attributes(self):
        pass


class TestSpecialDeriveField(unittest.TestCase):

    def test_ctor(self):
        pass

    def test_attributes(self):
        pass


class TestSpecialDeriveEnumeratedValues(unittest.TestCase):

    def test_ctor(self):
        pass

    def test_attributes(self):
        pass


class TestSpecialDeriveAll(unittest.TestCase):

    def test_ctor(self):
        pass

    def test_attributes(self):
        pass
